"""
hydro_bubbles.py is a Python routine that contains functions
to study the 1D hydrodynamic solutions of expanding bubbles
produced from first-order phase transitions.

Currently part of the cosmoGW code:

https://github.com/cosmoGW/cosmoGW/
https://github.com/cosmoGW/cosmoGW/blob/main/src/cosmoGW/hydro_bubbles.py

Author: Alberto Roper Pol
Created: 01/02/2023
Updated: 03/06/2025 (release cosmoGW 1.0: https://pypi.org/project/cosmoGW)

Other contributors: Antonino Midiri, Simona Procacci

Main references are:

Appendix A of RoperPol:2025a - A. Roper Pol, S. Procacci, A. S. Midiri,
C. Caprini, "Irrotational fluid perturbations from first-order phase
transitions," in preparation

Other relevant references are:

Espinosa:2010hh - J. R. Espinosa, T. Konstandin, J. M. No, G. Servant,
"Energy Budget of Cosmological First-order Phase Transitions,"
JCAP 06 (2010) 028, arXiv:1004.4187

Hindmarsh:2016lnk - M. Hindmarsh, "Sound shell model for acoustic gravitational
wave production at a first-order phase transition in the early Universe,"
Phys. Rev. Lett. 120 (2018) 7, 071301, arXiv:1608.04735

Hindmarsh:2019phv - M. Hindmarsh, M. Hijazi, "Gravitational waves from first order
cosmological phase transitions in the Sound Shell Model,"
JCAP 12 (2019) 062, arXiv:1909.10040

RoperPol:2023dzg - A. Roper Pol, S. Procacci, C. Caprini,
"Characterization of the gravitational wave spectrum from sound waves within
the sound shell model," Phys. Rev. D 109, 063531 (2024), arXiv:2308.12943.
"""

import numpy  as np
import pandas as pd
import matplotlib.pyplot as plt

# reference values
cs2_ref   = 1/3        # speed of sound squared
Nxi_ref   = 10000      # reference discretization in xi
Nxi2_ref  = 10         # reference discretization in xi out
                       # of the profiles
Nvws_ref  = 20         # reference discretization in vwall
tol_ref   = 1e-5       # reference tolerance on shooting algorithm
it_ref    = 30         # reference number of iterations

# reference values for a deflagration
vw_def    = 0.5
alpha_def = 0.263
# reference values for a hybrid
vw_hyb    = 0.7
alpha_hyb = 0.052
# reference values for a detonation
vw_det    = 0.77
alpha_det = 0.091

# reference set of colors
cols_ref = ['black', 'darkblue', 'blue', 'darkgreen', 'green',
            'purple', 'darkred', 'red', 'darkorange', 'orange', 'violet']

def Chapman_Jouget(alp):

    '''
    Chapman-Jouget is the wall velocity at which the relative speed
    behind the wall becomes that of the spped of sound.
    It corresponds to the limiting case separating detonations
    and supersonic deflagrations.

    Reference: Eq. 39 of Espinosa:2010hh
    Note that Eq. B19 of Hindmarsh:2019phv has a typo

    Arguments:
        alp -- alpha at the + side of the wall (alpha_pl)
    Returns:
        vcJ -- Chapman-Jouget speed
    '''

    return 1./np.sqrt(3)/(1 + alp)*(1 + np.sqrt(alp*(2 + 3*alp)))

def type_nucleation(vw, alp, cs2=cs2_ref):

    '''
    Function that determines the type of bubble solution:
    a) subsonic deflagrations ('def'),
    b) supersonic deflagrations ('hyb'), or
    c) detonations ('det').

    Arguments:
        vw -- bubble wall speed
        alp -- strenght of the phase transition at the + side
                 of the wall (alpha_pl)
        cs2 -- square of the speed of sound (default 1/3)

    Returns:
        ty -- type of solution ('def', 'hyb', 'det')
    '''

    mult_vw  = isinstance(vw,  (list, tuple, np.ndarray))
    mult_alp = isinstance(alp, (list, tuple, np.ndarray))
    cs       = np.sqrt(cs2)

    if not mult_vw:  vw  = np.array([vw])
    if not mult_alp: alp = np.array([alp])
    v_cj = Chapman_Jouget(alp)

    ty = np.full((len(vw), len(alp)), 'hyb')
    vw, v_cj = np.meshgrid(vw, v_cj, indexing='ij')
    ty[vw < cs]   = 'def'
    ty[vw > v_cj] = 'det'

    if   not mult_alp and not mult_vw: ty = ty[0, 0]
    elif not mult_alp: ty = ty[:, 0]
    elif not mult_vw:  ty = ty[0, :]

    return ty

######## 1D HYDRO SOLUTIONS UNDER SPHERICAL SYMMETRY ########

'''
The solutions are based in the 1D hydrodynamic descriptions
of Espinosa:2010hh and Hindmarsh:2019phv.

The details accompanying the code are provided in RoperPol:2023dzg
and RoperPol:2025a (appendix A)
'''

def Lor_mu(v, vw):

    '''
    Lorentz transform of the velocity v in the reference frame of the wall
    (with speed vw)

    mu(vw, v) = (vw - v)/(1 - vw*v)

    Reference: Eq. B12 of Hindmarsh:2019phv
    '''

    return (vw - v)/(1 - v*vw)

### hydrodynamic shocks with no vacuum energy

def v_shock(xi):

    '''
    Function that computes the shock velocity at the - side of the
    shock. Computed from the v_+ v_- = 1/3 condition.
    '''

    vsh = (3*xi**2 - 1)/2/xi

    return vsh

def w_shock(xi):

    '''
    Function that computes the ratio of enthalpies w-/w+ across the
    shock. Computed from the v_+ v_- = 1/3 condition.
    '''

    wsh = (9*xi**2 - 1)/3/(1 - xi**2)

    return wsh

#### differential equation for the velocity radial profile

def xi_o_v(xi, v, cs2=cs2_ref):

    '''
    Function that characterizes the 1d hydro equation under radial
    symmetry.

    It returns the value of dxi/dv used to solve the
    equation in compute_xi_from_v using RK4.

    Reference: Eq. 27 of Espinosa:2010hh

    Arguments:
        xi  -- self-similar r/t
        v   -- 1d velocity profile
        cs2 -- square of the speed of sound (default 1/3)

    Returns:
        f   -- value of f(xi, v) = dxi/dv
    '''

    gamma2 = 1/(1 - v**2)
    mu = Lor_mu(v, xi)
    f = xi*gamma2*(1 - xi*v)*(mu**2/cs2 - 1)/2/v

    return f

def compute_xi_from_v(v, xi0, cs2=cs2_ref, shock=False):

    '''
    Function that computes the solution xi (v) using a 4th-order
    Runge-Kutta scheme. Since dv/dxi has a singularity, it is necessary
    to compute xi(v) and then invert each of the solutions to have the full
    dynamical solution. However, for the physical solution, computing v (xi)
    is more practical.

    Arguments:
        v     -- velocity array
        xi0   -- position where boundary condition is known (it has to correspond
                 to the first value in the velocity array)
        cs2   -- square of the speed of sound (default 1/3)
        shock -- option to stop the integration when a shock is found
                 (as happens in deflagrations)

    Returns:
        xi    -- self-similar array
        sh    -- boolean determining if a shock is formed
        indsh -- index determining the position of the shock
                 (-1 if no shock)
    '''

    xi = np.zeros(len(v)) + xi0
    sh = False
    indsh = -1

    for i in range(0, len(v) - 1):
        dv = v[i+1] - v[i]
        k1 = xi_o_v(xi[i], v[i])
        k2 = xi_o_v(xi[i] + dv*k1/2, .5*(v[i+1] + v[i]))
        k3 = xi_o_v(xi[i] + dv*k2/2, .5*(v[i+1] + v[i]))
        k4 = xi_o_v(xi[i] + dv*k3, v[i + 1])
        xi_new = xi[i] + 1/6*(k1 + 2*k2 + 2*k3 + k4)*dv
        if shock:
            xi_sh = xi_new
            v_sh = v_shock(xi_sh)
            if v[i + 1] < v_sh:
                xi[i + 1:] = xi_sh
                sh = True
                indsh = i
                break
        xi[i + 1] = xi_new
        if xi_new > 1: xi[i + 1] = 1

    return xi, sh, indsh

def compute_int_w(xi, v, cs2=cs2_ref):

    '''
    Function that computes the integrand for the integration of dw/dxi
    (enthalpy) equation, as a function of the solution v(xi).

    Reference: Eq. (29) of Espinosa:2010hh

     Arguments:
        xi  -- self-similar r/t
        v   -- 1d velocity profile
        cs2 -- square of the speed of sound (default 1/3)

    '''

    return (1. + 1./cs2)/(1. - v**2)*Lor_mu(v, xi)

def compute_w(v, xi, cs2=cs2_ref):

    '''
    Function that computes the enthalpy from the solution of the
    velocity profile.

     Arguments:
        xi  -- self-similar r/t
        v   -- 1d velocity profile
        cs2 -- square of the speed of sound (default 1/3)
    '''

    w = np.zeros(len(v)) + 1
    ss = 0
    for i in range(0, len(v) - 1):
        ff_ip = compute_int_w(xi[i + 1], v[i + 1], cs2=cs2)
        ff_i = compute_int_w(xi[i], v[i], cs2=cs2)
        ss += .5*(v[i + 1] - v[i])*(ff_ip + ff_i)
        w[i + 1] = np.exp(ss)

    return w

######### SOLVE FOR THE DIFFERENT TYPE OF BOUNDARY CONDITIONS #########

############# MATCHING CONDITIONS ACROSS DISCONTINUITIES ##############

def vp_tilde_from_vm_tilde(vw, alpha, plus=True, sg='plus'):

    '''
    Function that computes the + (symmetric phase) or - (broken phase)
    velocity, defined in the wall reference frame, across the wall,
    as a function of the value of the velocity at the opposite side of
    the wall (via the matching conditions).

    Reference: Eqs. B6 and B7 of Hindmarsh:2019phv

    Arguments:
        vw    -- velocity imposed at one of the sides of the wall.
                 It usually is the wall velocity but becomes cs for
                 hybrids
        alpha -- phase transition strength at the symmetric phase
                 (can be different than that at the nucleation temperature)
        plus  -- option to consider positive (True) or negative (False)
                 branch of the equation found from the matching conditions
                 (default is True)
        sg    -- option to compute v+ from v- = vw if sg == 'plus' or to compute
                 v- from v+ = vw if sg == 'minus' (default is 'plus')

    Returns:
        vp_vm -- v+ or v- from the value at the other side of the bubble wall
    '''

    if sg == 'plus':

        a1 = 1/3/vw + vw
        a2 = np.sqrt((1/3/vw - vw)**2 + 4*alpha**2 + 8/3*alpha)
        aa = .5/(1 + alpha)

    else:

        a1 = (1 + alpha)*vw + (1 - 3*alpha)/3/vw
        a2 = np.sqrt(((1 + alpha)*vw + (1 - 3*alpha)/3/vw)**2 - 4/3)
        aa = 1/2

    if plus: a = a1 + a2
    else: a = a1 - a2

    vp_vm = aa*a

    return vp_vm

def vplus_vminus(alpha, vw=1., ty='det', cs2=cs2_ref):

    '''
    Function that returns v_+ and v_- (in the wall frame)
    for the different type of solutions
    (deflagrations, detonations and hybrids).

    This allows to give the boundary conditions corresponding to each
    of the solutions.

    Reference: Appendix A of RoperPol:2025a

    Arguments:
        alpha -- value of alpha at the + side of the wall
        vw    -- wall velocity
        cs2   -- speed of sound squared (default is 1/3)
        ty    -- type of solution

    Returns:
        vplus, vminus -- + and - velocities expressed in the wall
                            reference of frame
    '''

    cs = np.sqrt(cs2)

    if not isinstance(ty, (list, tuple, np.ndarray)):

        if ty == 'det':
            vplus = vw
            vminus = vp_tilde_from_vm_tilde(vw, alpha, plus=True, sg='minus')
        if ty == 'def':
            vminus = vw
            vplus = vp_tilde_from_vm_tilde(vw, alpha, plus=False, sg='plus')
        if ty == 'hyb':
            vminus = cs
            vplus = vp_tilde_from_vm_tilde(cs, alpha, plus=False, sg='plus')

    else:

        vplus    = np.zeros(len(vw))
        vminus   = np.zeros(len(vw))
        inds_det = np.where(ty == 'det')
        inds_def = np.where(ty == 'def')
        inds_hyb = np.where(ty == 'hyb')
        vplus[inds_det]  = vw[inds_det]
        vminus[inds_det] = vp_tilde_from_vm_tilde(vw, alpha,
                                    plus=True, sg='minus')[inds_det]
        vplus[inds_def]  = vp_tilde_from_vm_tilde(vw, alpha,
                                    plus=False, sg='plus')[inds_def]
        vminus[inds_def] = vw[inds_def]
        vplus[inds_hyb]  = vp_tilde_from_vm_tilde(cs, alpha,
                                    plus=False, sg='plus')
        vminus[inds_hyb] = cs

    return vplus, vminus

######## function that computes the detonation part of the solutions ########
def det_sol(v0, xi0, cs2=cs2_ref, Nxi=Nxi_ref, zero_v=-4):

    '''
    Function that computes a detonation solution with boundary condition
    v0 at xi0

    Arguments:
        v0     -- value of v0 at the boundary
        xi0    -- position of the boundary
        cs2    -- speed of sound squared (default is 1/3)
        Nxi    -- number of discretization points in xi
        zero_v -- reference zero velocity (default 1e-4)

    Returns:
        xis    -- array of xi
        vs     -- array of 1d velocities
        ws     -- array of 1d enthalpies
    '''

    # compute solution from initial condition v = v0 at xi = vw
    # until v reduces to 4 orders of magntiude below value v minus
    cs = np.sqrt(cs2)

    # included option to initialize with multiple vws
    if not isinstance(xi0, (list, tuple, np.ndarray)):

        vs = np.logspace(np.log10(v0), np.log10(v0) + zero_v, Nxi)
        xis, sh, indsh = compute_xi_from_v(vs, xi0, cs2=cs2, shock=False)
        xi_sh = xis[indsh]
        ws    = compute_w(vs, xis, cs2=cs2)

        inds_sort = np.argsort(xis)
        xis = xis[inds_sort]
        vs  = vs[inds_sort]
        ws  = ws[inds_sort]

    else:

        xis = np.zeros((len(xi0), Nxi))
        vs  = np.zeros((len(xi0), Nxi))
        ws  = np.zeros((len(xi0), Nxi))

        for i in range(0, len(xi0)):

            vs[i, :]  = np.logspace(np.log10(v0), np.log10(v0) + zero_v, Nxi)
            xis[i, :], _, _ = compute_xi_from_v(vs[i, :], xi0[i], cs2=cs2,
                                                shock=False)
            ws[i, :]  = compute_w(vs[i, :], xis[i, :], cs2=cs2)

            inds_sort = np.argsort(xis[i, :])
            xis[i, :] = xis[i, inds_sort]
            vs[i, :]  = vs[i, inds_sort]
            ws[i, :]  = ws[i, inds_sort]

    return xis, vs, ws

####### function that computes the deflagration part of the solutions #######
def def_sol(v0, xi0, cs2=cs2_ref, Nxi=Nxi_ref, shock=True, zero_v=-4):

    '''
    Function that computes a deflagration solution with boundary condition
    v0 at xi0

    Arguments:
        v0     -- value of v0 at the boundary
        xi0    -- position of the boundary
        cs2    -- speed of sound squared (default is 1/3)
        Nxi    -- number of discretization points in xi
        shock  -- possibility to stop the calculation once a shock is formed
                 (default is True)
        zero_v -- reference zero velocity (default 1e-4)

    Returns:
        xis    -- array of xi
        vs     -- array of 1d velocities
        ws     -- array of 1d enthalpies
        xi_sh, sh -- position of the shock and boolean which becomes
                        True if a shock forms
    '''

    vs = np.logspace(np.log10(v0), np.log10(v0) + zero_v, Nxi)
    cs = np.sqrt(cs2)
    xi_sh = cs
    sh    = False
    v_sh  = 0

    if shock:
        xiss, sh, indsh = compute_xi_from_v(vs, xi0, cs2=cs2, shock=True)
        xi_sh = xiss[indsh]
        if not sh: xi_sh = cs
        v_sh  = v_shock(xi_sh)
        xis   = np.linspace(xi0, xi_sh, Nxi + 1)
        xis   = xis[:Nxi-1]
        vs    = np.interp(xis, xiss, vs)
        vs    = np.append(vs, v_sh)
        xis   = np.append(xis, xi_sh)

    else:
        xis, sh, indsh = compute_xi_from_v(vs, xi0, cs2=cs2, shock=False)
        xi_sh = xis[indsh]

    ws   = compute_w(vs, xis, cs2=cs2)
    w_sh = w_shock(xi_sh)
    ws   = ws*w_sh/ws[-1]

    return xis, vs, ws, xi_sh, sh

def compute_def(vw=vw_def, alpha=alpha_def, cs2=cs2_ref, Nxi=Nxi_ref,
                shock=True):

    '''
    Function that computes the solutions for a subsonic deflagration
    1d profile given vw and alpha, using def_sol

    Arguments:
        vw    -- wall velocity (default is 0.5)
        alpha -- strength of the phase transition (default is 0.263)
        cs2   -- speed of sound squared (default is 1/3)
        Nxi   -- number of discretization points in xi
        shock -- possibility to stop the calculation once a shock is formed
                 (default is True)

    Returns:
        xis   -- array of xi
        vs    -- array of 1d velocities
        ws    -- array of 1d enthalpies
        xi_sh, sh -- position of the shock and boolean which becomes
                        True if a shock forms
        w_pl, w_m -- plus and minus values of the enthalpies
                        across the bubble
    '''

    ## relative velocity at + is computed from \tilde v- = \xi_w
    vrels, _ = vplus_vminus(alpha, vw=vw, ty='def')
    # Lorentz boosted v plus
    vpl = Lor_mu(vrels, vw)

    xis, vs, ws, xi_sh, sh = def_sol(vpl, vw, cs2=cs2, Nxi=Nxi, shock=shock)

    # values at both sides of the bubble wall
    w_pl = ws[0]
    w_m = w_pl*vrels/(1 - vrels**2)/vw*(1 - vw**2)

    return xis, vs, ws, xi_sh, sh, w_pl, w_m

def compute_hyb(vw=vw_hyb, alpha=alpha_hyb, cs2=cs2_ref,Nxi=Nxi_ref,
                shock=True):

    '''
    Function that computes the solutions for a supersonic deflagration
    1d profile given vw and alpha, using det_sol and def_sol

    Arguments:
        vw    -- wall velocity (default is 0.7)
        alpha -- strength of the phase transition (default is 0.052)
        cs2   -- speed of sound squared (default is 1/3)
        Nxi   -- number of discretization points in xi
        shock -- possibility to stop the calculation once a shock is formed
                 (default is True)

    Returns:
        xis   -- array of xi
        vs    -- array of 1d velocities
        ws    -- array of 1d enthalpies
        xi_sh, sh -- position of the shock and boolean which becomes
                        True if a shock forms
        w_pl, w_m -- plus and minus values of the enthalpies
    '''

    cs = np.sqrt(cs2)

    ## relative velocity at + is computed from \tilde v- = cs
    vrels, _ = vplus_vminus(alpha, cs2=cs2, ty='hyb')
    vpl = Lor_mu(vrels, vw)
    vm  = Lor_mu(cs, vw)

    # compute deflagration solution
    xis, vs, ws, xi_sh, sh = def_sol(vpl, vw, cs2=cs2,
                                     Nxi=int(Nxi/2), shock=shock)

    # compute detonation solution
    xis2, vs2, ws2 = det_sol(vm, vw, cs2=cs2, Nxi=int(Nxi/2))
    # ratio of w+ over w- across the bubble wall
    w_pl = ws[0]
    w_m  = w_pl*vrels/(1 - vrels**2)*(1 - cs2)/cs
    ws2 *= w_m

    xis = np.append(xis2, xis)
    vs  = np.append(vs2, vs)
    ws  = np.append(ws2, ws)

    return xis, vs, ws, xi_sh, sh, w_pl, w_m

def compute_det(vw=vw_det, alpha=alpha_det, cs2=cs2_ref, Nxi=Nxi_ref):

    '''
    Function that computes the solutions for a detonation 1d profile
    given vw and alpha, using det_sol

    Arguments:
        vw    -- wall velocity (default is 0.77)
        alpha -- strength of the phase transition (default is 0.091)
        cs2   -- speed of sound squared (default is 1/3)
        Nxi   -- number of discretization points in xi

    Returns:
        xis   -- array of xi
        vs    -- array of 1d velocities
        ws    -- array of 1d enthalpies
        w_pl, w_m -- plus and minus values of the enthalpies
                        across the bubble
    '''

    ## relative velocity at - is computed from \tilde v+ = \xi_w
    _, vrels = vplus_vminus(alpha, vw=vw, ty='det')
    # Lorentz boosted v minus
    vm   = Lor_mu(vrels, vw)
    w_m  = vw/(1 - vw**2)/vrels*(1 - vrels**2)
    w_pl = 1

    xis, vs, ws = det_sol(vm, vw, cs2=cs2, Nxi=Nxi)
    ws *= w_m

    # no shock is formed in detonations, so xi_sh is set to vw
    # and sh to False
    xi_sh = vw
    sh    = False

    return xis, vs, ws, xi_sh, sh, w_pl, w_m

def compute_alphan(vw=vw_def, alpha_obj=alpha_def, tol=tol_ref, cs2=cs2_ref,
                   quiet=False, max_it=it_ref, meth=1, Nxi=Nxi_ref, ty='def'):

    '''
    Function that computes the value of \alpha_+ corresponding to alpha.
    It requires to compute the 1d profile of w and then iteratively look for
    the value of alpha_+ that gives the correct alpha.

    Arguments:
        vw        -- wall velocity
        alpha_obj -- value of alpha defined at the nucleation temperature that
                        wants to be reached
        tol       -- relative tolerance to consider convergence (default is 1e-4)
        cs2       -- speed of sound squared (default is 1/3)
        quiet     -- option to avoid printing some debug options (default is False)
        max_it    -- maximum number of allowed iterations (default is 30)
        meth      -- method on the Newton-Raphson update (2 options)
        Nxi       -- number of points in xi discretization (default is 1000)
        ty        -- type of solution (options are def or hyb)

    Returns:
        xis0, vvs0, wws0 -- arrays of xi, velocity and enthalpy of the converged
                               solutions
        xi_sh, sh, w_pl, w_m -- shock position, boolean if shock has formed,
                                    plus and minus enthalpies
        alpha_n   -- converged (or not) alpha at nucleation
        alp_plus  -- value of alpha_+ leading to alpha_obj
        conv      -- boolean determining if the algorithm has converged
    '''

    # first guess
    alp_plus = alpha_obj

    j = 0
    conv = False

    while not conv and j < max_it:

        j += 1

        if ty == 'hyb':
            xis0, vvs0, wws0, xi_sh, sh, w_pl, w_m = \
                compute_hyb(vw=vw, alpha=alp_plus, cs2=cs2, Nxi=Nxi, shock=True)

        if ty == 'def':
            xis0, vvs0, wws0, xi_sh, sh, w_pl, w_m = \
                compute_def(vw=vw, alpha=alp_plus, cs2=cs2, Nxi=Nxi, shock=True)

        alpha_n = alp_plus*w_pl

        if abs(alpha_n - alpha_obj)/alpha_obj < tol: conv = True

        else:

            if meth==1: alp_plus = alpha_obj/w_pl
            if meth==2: alp_plus += (alpha_obj - alpha_n)/w_pl

        if not quiet:
            print('iteration', j, 'alpha', alpha_n)
            print('iteration', j, 'new guess', alp_plus)

    if not quiet:
        print(j, 'iterations for vw=', vw,' and alpha= ',alpha_obj)
        print('alpha:', alpha_n, ', alpha_+:', alp_plus)

    return xis0, vvs0, wws0, xi_sh, sh, w_pl, w_m, alpha_n, alp_plus, conv

def compute_profiles_vws_multalp(alphas, vws=[0], cs2=cs2_ref, Nvws=Nvws_ref, Nxi=Nxi_ref,
                                 Nxi2=Nxi2_ref, alphan=True, quiet=True, tol=tol_ref,
                                 max_it=it_ref, lam=False, eff=False):

    '''
    Function that computes the velocity and enthalpy profiles for an array
    of alpha using compute_profiles_vws function

    Arguments:
        alphas -- nucleation T alphas
        vws    -- range of wall velocities
        cs2    -- square of the speed of sound (default is 1/3)
        Nvws   -- number of wall velocities (if vws is not given)
        Nxi    -- number of discretization points in xi where the profiles are
                  computed
        Nxi2   -- number of discretization points in xi out of the solution
                  of the 1d profiles
        alphan -- option to identify if the input alpha is at the
                  nucleation temperature (if True) or alpha+ (if False),
                  default is True
        quiet  -- option to avoid printing debugging information (default is True)
        max_it -- maximum number of iterations to find alpha_+
        lam    -- option to compute energy perturbations lambda instead
                  of enthalpy (default is False, so enthalpy)
        tol    -- tolerance of the relative error to consider convergence
                  of alpha_+ has been reached
        eff    -- option to compute the efficiency factors kappa and
                  omega (default is False)

    Returns:
        xis       -- array of xi positions
        vvs       -- array of velocities
        wws       -- array of enthalpies (energy density perturbations
                     if lam is True)
        alphas_n  -- array of nucleation alphas (if input is alpha+, when
                     alphan is False, otherwise it returns array of values
                     of alpha+)
        conv      -- array of booleans determining if the alphan calculation
                     has converged
        shocks    -- array of booleans determining if a shock is produced
                     in the fluid profile
        xi_shocks -- positions of the shocks (if shock is not produced,
                     it returns xi_front)
        wms       -- enthalpy values at - side of the wall
        kappas    -- ratio of kinetic to vacuum energy density
                     (only returned if eff is True)
        omegas    -- ratio of energy density perturbations to vacuum
                     energy density (only returned if eff is True)
    '''

    if len(np.shape(vws)) == 1:
        if len(vws) == 0: vws = np.linspace(0.1, .99, Nvws)
    xis = np.linspace(0, 1, Nxi + Nxi2)

    if isinstance(alphas, (list, tuple, np.ndarray)):

        vvs       = np.zeros((len(vws), len(alphas), len(xis)))
        wws       = np.zeros((len(vws), len(alphas), len(xis))) + 1.
        alphas_n  = np.zeros((len(vws), len(alphas)))
        conv      = np.zeros((len(vws), len(alphas))) + 1.
        shocks    = np.zeros((len(vws), len(alphas)))
        xi_shocks = np.zeros((len(vws), len(alphas)))
        wms       = np.zeros((len(vws), len(alphas)))

    if eff:

        if isinstance(alphas, (list, tuple, np.ndarray)):

            kappas = np.zeros((len(vws), len(alphas)))
            omegas = np.zeros((len(vws), len(alphas)))

            for i in range(0, len(alphas)):

                if not quiet:
                    print('Computing alpha = %'%alphas[i], ' out of',
                          '%i'%(len(alphas) + 1))

                xis, vvs[:, i, :], wws[:, i, :], alphas_n[:, i], \
                    conv[:, i], shocks[:, i], xi_shocks[:, i], wms[:, i], \
                    kappas[:, i], omegas[:, i] = \
                        compute_profiles_vws(alphas[i], vws=vws, cs2=cs2,
                            Nxi=Nxi, Nxi2=Nxi2, plot=False, alphan=alphan,
                            quiet=True, tol=tol, max_it=max_it, lam=lam,
                            eff=eff)

        else:

            xis, vvs, wws, alphas_n, conv, shocks, xi_shocks, wms, \
                    kappas, omegas = \
                        compute_profiles_vws(alphas, vws=vws, cs2=cs2,
                            Nxi=Nxi, Nxi2=Nxi2, plot=False, alphan=alphan,
                            quiet=True, tol=tol, max_it=max_it, lam=lam,
                            eff=eff)

        return xis, vvs, wws, alphas_n, conv, shocks, xi_shocks, wms, kappas, omegas

    else:

        if isinstance(alphas, (list, tuple, np.ndarray)):

            for i in range(0, len(alphas)):

                if not quiet:
                    print('Computing alpha = %f'%alphas[i], ' out of',
                          '%i'%(len(alphas) + 1))

                xis, vvs[:, i, :], wws[:, i, :], alphas_n[:, i], \
                    conv[:, i], shocks[:, i], xi_shocks[:, i], wms[:, i] = \
                        compute_profiles_vws(alphas[i], vws=vws, cs2=cs2,
                            Nxi=Nxi, Nxi2=Nxi2, plot=False, alphan=alphan,
                            quiet=True, tol=tol, max_it=max_it, lam=lam,
                            eff=eff)

        else:

            xis, vvs, wws, alphas_n, conv, shocks, xi_shocks, wms = \
                compute_profiles_vws(alphas, vws=vws, cs2=cs2,
                                     Nxi=Nxi, Nxi2=Nxi2, plot=False, alphan=alphan,
                                     quiet=True, tol=tol, max_it=max_it, lam=lam,
                                     eff=eff)

        return xis, vvs, wws, alphas_n, conv, shocks, xi_shocks, wms

def compute_profiles_vws(alpha, vws=[], cs2=cs2_ref, Nvws=Nvws_ref, Nxi=Nxi_ref,
                         Nxi2=Nxi2_ref, plot=False, plot_v='v', cols=[],
                         alphan=True, quiet=True, tol=tol_ref, max_it=it_ref,
                         ls='solid', alp=1., lam=False, meth=1, legs=False,
                         fs_lg=14, st_lg=2, eff=False, save=False, dec_vw=1,
                         ress='results/1d_profiles', strs_vws=[], str_alp=[]):

    '''
    Function that computes the velocity and enthalpy profiles for a given
    alpha (at nucleation T, not alpha+) and a range of wall velocities.

    Arguments:
        alpha  -- nucleation T alpha
        vws    -- range of wall velocities
        cs2    -- square of the speed of sound (default is 1/3)
        Nvws   -- number of wall velocities (if vws is not given)
        Nxi    -- number of discretization points in xi where the profiles are
                  computed
        Nxi2   -- number of discretization points in xi out of the solution
                  of the 1d profiles
        plot   -- option to plot the resulting 1d profiles
        plot_v -- choice of plotting ('v' for velocity, 'w' for
                  enthalpy, 'both' for both)
        cols   -- array with colors (cols_ref by default)
        alphan -- option to identify if the input alpha is at the
                  nucleation temperature (if True) or alpha+ (if False),
                  default is True
        quiet  -- option to avoid printing debugging information (default is True)
        max_it -- maximum number of iterations to find alpha_+
        ls     -- line styles
        alp    -- opacity of the plots
        lam    -- option to compute energy perturbations lambda instead
                  of enthalpy (default is False, so enthalpy)
        tol    -- tolerance of the relative error to consider convergence
                  of alpha_+ has been reached
        legs   -- legends to be included (default is False)
        save   -- option to save in a file the results of the 1d
                  profiles (default is False)
        ress   -- directory where to save the files (default is
                  'results/1d_profiles')
        eff    -- option to compute the efficiency factors kappa and
                  omega (default is False)

    Returns:
        xis       -- array of xi positions
        vvs       -- array of velocities
        wws       -- array of enthalpies (energy density perturbations
                     if lam is True)
        alphas_n  -- array of nucleation alphas (if input is alpha+, when
                     alphan is False, otherwise it returns array of values
                     of alpha+)
        conv      -- array of booleans determining if the alphan calculation
                     has converged
        shocks    -- array of booleans determining if a shock is produced
                     in the fluid profile
        xi_shocks -- positions of the shocks (if shock is not produced,
                     it returns xi_front)
        wms       -- enthalpy values at - side of the wall
        kappas    -- ratio of kinetic to vacuum energy density
                     (only returned if eff is True)
        omegas    -- ratio of energy density perturbations to vacuum
                     energy density (only returned if eff is True)
    '''

    if len(vws) == 0: vws = np.linspace(0.1, .99, Nvws)
    vCJ       = Chapman_Jouget(alp=alpha)
    cs        = np.sqrt(cs2)
    xis       = np.linspace(0, 1, Nxi + Nxi2)
    vvs       = np.zeros((len(vws), len(xis)))
    wws       = np.zeros((len(vws), len(xis))) + 1.
    alphas_n  = np.zeros(len(vws))
    conv      = np.zeros(len(vws)) + 1.
    kappas    = np.zeros(len(vws))
    omegas    = np.zeros(len(vws))
    shocks    = np.zeros(len(vws))
    xi_shocks = np.zeros(len(vws))
    wms       = np.zeros(len(vws))

    if plot_v == 'both' and plot:
        plt.figure(1)
        plt.figure(2)

    if len(cols) == 0: cols = cols_ref

    for i in range(0, len(vws)):

        # determine type of solution
        ty = type_nucleation(vws[i], alpha, cs2=cs2)

        if ty == 'def':

            ## iteratively compute the real alpha_+ leading to alpha
            if alphan:
                xis0, vvs0, wws0, xi_sh, sh, w_pl, w_m, alpha_n, \
                    alp_plus, conv[i] = compute_alphan(vw=vws[i], alpha_obj=alpha,
                                            tol=tol, cs2=cs2, meth=meth,
                                            quiet=quiet, max_it=max_it, Nxi=Nxi,
                                            ty='def')
                alphas_n[i] = alp_plus

            ## otherwise, alpha given is assumed to be alpha_+
            else:
                xis0, vvs0, wws0, xi_sh, sh, w_pl, w_m = \
                        compute_def(vw=vws[i], alpha=alpha, cs2=cs2,
                                    Nxi=Nxi, shock=True)
                alphas_n[i] = alpha*w_pl

            inds = np.where((xis >= vws[i])*(xis <= xi_sh))[0]
            inds2 = np.where(xis < vws[i])[0]
            wws[i, inds2] = w_m

        elif ty == 'hyb':

            ## iteratively compute the real alpha_+ leading to alpha
            if alphan:
                xis0, vvs0, wws0, xi_sh, sh, w_pl, w_m, alpha_n, \
                    alp_plus, conv[i] = compute_alphan(vw=vws[i], alpha_obj=alpha,
                                            tol=tol, cs2=cs2, ty='hyb', meth=meth,
                                            quiet=quiet, max_it=max_it, Nxi=Nxi)
                alphas_n[i] = alp_plus

            ## otherwise, alpha given is assumed to be alpha_+
            else:
                xis0, vvs0, wws0, xi_sh, sh, w_pl, w_m = \
                        compute_hyb(vw=vws[i], alpha=alpha, cs2=cs2,
                                    Nxi=Nxi, shock=True)
                alphas_n[i] = alpha*w_pl

            inds = np.where((xis >= cs)*(xis <= xi_sh))[0]
            if xi_sh == vws[i]: inds = \
                    np.append(inds, np.where((xis <= xi_sh))[-1] + 1)
            inds2 = np.where(xis < cs)[0]
            wws[i, inds2] = wws0[0]

        else:

            xis0, vvs0, wws0, xi_sh, sh, w_pl, w_m = \
                    compute_det(vw=vws[i], alpha=alpha, cs2=cs2,
                                Nxi=Nxi)

            inds  = np.where((xis >= cs)*(xis <= vws[i]))[0]
            inds2 = np.where(xis < cs)[0]
            wws[i, inds2] = wws0[0]
            alphas_n[i]   = alpha

        vvs[i, inds] = np.interp(xis[inds], xis0, vvs0)
        wws[i, inds] = np.interp(xis[inds], xis0, wws0)
        shocks[i]    = sh
        xi_shocks[i] = xi_sh
        wms[i]       = w_m

        # compute efficiency of energy density production
        if eff:
            kappas[i], omegas[i] = kappas_from_prof(vws[i], alpha, xis,
                                                    wws[i, :], vvs[i, :])


        # compute mean energy density from enthalpy if lam is True
        if lam:
            if alphan: alp_lam = alpha
            else: alp_lam      = alphas_n[i]
            wws[i, :]          = w_to_lam(xis, wws[i, :], vws[i], alp_lam)

        if plot:

            j = i%11
            if st_lg == 1: str_lg = r'$\xi_w=%.1f$'%vws[i]
            if st_lg == 2: str_lg = r'$\xi_w=%.2f$'%vws[i]
            if plot_v=='v': plt.plot(xis, vvs[i, :], color=cols[j], ls=ls, alpha=alp,
                                     label=str_lg)
            if plot_v=='w': plt.plot(xis, wws[i, :], color=cols[j], ls=ls, alpha=alp,
                                     label=str_lg)
            if plot_v=='both':
                plt.figure(1)
                plt.plot(xis, vvs[i, :], color=cols[j], ls=ls, alpha=alp, label=str_lg)
                plt.figure(2)
                plt.plot(xis, wws[i, :], color=cols[j], ls=ls, alpha=alp, label=str_lg)

        if save and alphan:

            # option to save the results on files (one file for each alpha
            # and wall velocity)
            # save is only possible when the input alpha is that at the
            # nucleation temperature (alphan = True)
            df = pd.DataFrame({'alpha': alpha*xis**0, 'xi_w': vws[i]*xis**0,
                               'xi': xis, 'v': vvs[i, :], 'w': wws[i, :],
                               'alpha_pl': alphas_n[i]*xis**0, 'shock': shocks[i]*xis**0,
                               'xi_sh': xi_shocks[i]*xis**0, 'wm': wms[i]*xis**0})
            # save file
            if len(str_alp) == 0:
                str_alp = '%s'%(alpha)
                str_alp = '0' + str_alp
                str_alp = str_alp[2:]
            if len(strs_vws) == 0:
                str_vws = '%s'%(np.round(vws[i], decimals=dec_vw))
                str_vws = str_vws[2:]
            else: str_vws = strs_vws[i]
            file_dir = ress + '/alpha_%s_vw_%s.csv'%(str_alp, str_vws)
            try:
                df.to_csv(file_dir)
                print('results of 1d profile saved in ', file_dir)
            except:
                print('create directory results/1d_profiles to save',
                      ' the 1d profiles')

    if plot:

        if plot_v == 'v' or plot_v == 'both':
            if plot_v == 'both': plt.figure(1)
            plt.ylim(-.05, 1.05)
            plt.ylabel(r'$ v_{\rm ip} (\xi)$')
        if plot_v == 'w' or plot_v == 'both':
            if plot_v == 'both': plt.figure(2)
            plt.ylim(0, 5)
            if lam: plt.ylabel(r'$ \lambda_{\rm ip} (\xi)$')
            else: plt.ylabel(r'$ w(\xi)$')
        l = [1]
        if plot_v == 'both': l = [1, 2]
        for j in l:
            plt.figure(j)
            plt.xlim(0, 1)
            plt.vlines(cs, -5, 30, color='black', ls='dashed', lw=1)
            plt.vlines(vCJ, -5, 30, color='black', ls='dashed', lw=1)
            plt.xlabel(r'$\xi$')
            if legs: plt.legend(fontsize=fs_lg)

    if eff:
        return xis, vvs, wws, alphas_n, conv, shocks, xi_shocks, wms, kappas, omegas
    else:
        return xis, vvs, wws, alphas_n, conv, shocks, xi_shocks, wms

################### COMPUTING EFFICIENCIES FROM 1D PROFILES ###################

def kappas_from_prof(vw, alpha, xis, ws, vs):

    '''
    Function that computes the kinetic energy density efficiency kappa
    and thermal factor omega from the 1d profiles.
    '''

    try:
        kappa = 4/vw**3/alpha*np.trapezoid(xis**2*ws/(1 - vs**2)*vs**2, xis)
        omega = 3/vw**3/alpha*np.trapezoid(xis**2*(ws - 1), xis)
    except:
        kappa = 4/vw**3/alpha*np.trapz(xis**2*ws/(1 - vs**2)*vs**2, xis)
        omega = 3/vw**3/alpha*np.trapz(xis**2*(ws - 1), xis)

    return kappa, omega

def kappas_Esp(vw, alp, cs2=cs2_ref):

    """"
    Function that computes the efficiency in converting vacuum to
    kinetic energy density for detonations, deflagrations and hybrids.

    Uses the semiempirical fits from Espinosa:2010hh, appendix A,
    following the bag equation of state.

    Numerical values can be computed from the 1d profiles using
    compute_profiles_vws function with eff = True

    Arguments:
        vw    -- value or array of wall velocities
        alp   -- strength of the phase transition at the nucleation temperature
        cs2   -- square of the speed of sound (default is 1/3)

    Returns:
        kappa -- ratio of kinetic to vacuum energy density, computed
                 in the bag equation of state
    """

    cs   = np.sqrt(cs2)

    mult_alp = isinstance(alp, (list, tuple, np.ndarray))
    mult_vw  = isinstance(vw,  (list, tuple, np.ndarray))

    if not mult_vw:  vw  = np.array([vw])
    if not mult_alp: alp = np.array([alp])
    v_cj = Chapman_Jouget(alp)

    ty       = type_nucleation(vw, alp, cs2=cs2)
    _,  v_cj = np.meshgrid(vw, v_cj, indexing='ij')
    vw, alp  = np.meshgrid(vw, alp,  indexing='ij')
    kappa    = np.zeros_like(vw)

    # kappa at vw << cs
    kapA = vw**(6/5)*6.9*alp/(1.36 - 0.037*np.sqrt(alp) + alp)
    # kappa at vw = cs
    kapB = alp**(2/5)/(0.017 + (0.997 + alp)**(2/5))
    # kappa at vw = cJ (Chapman-Jouget)
    kapC = np.sqrt(alp)/(0.135 + np.sqrt(0.98 + alp))
    # kappa at vw -> 1
    kapD = alp/(0.73 + 0.083*np.sqrt(alp) + alp)

    # deflagrations
    den       = (cs**(11/5) - vw**(11/5))*kapB + vw*cs**(6/5)*kapA
    kappa_def = cs**(11/5)*kapA*kapB/den
    kappa[ty == 'def'] = kappa_def[ty == 'def']

    # detonations
    kappa_det = (v_cj - 1)**3*(v_cj/vw)**(5/2)*kapC*kapD
    den       = ((v_cj - 1)**3 - (vw - 1)**3)*v_cj**(5/2)*kapC + (vw - 1)**3*kapD
    kappa_det = kappa_det/den
    kappa[ty == 'det'] = kappa_det[ty == 'det']

    # hybrids
    ddk        = -.9*np.log(np.sqrt(alp)/(1 + np.sqrt(alp)))
    kappa_hyb  = kapB + (vw - cs)*ddk
    kappa_hyb += ((vw - cs)/(v_cj - cs))**3*(kapC - kapB - (v_cj - cs)*ddk)
    kappa[ty == 'hyb'] = kappa_hyb[ty == 'hyb']

    if   not mult_alp and not mult_vw: kappa = kappa[0, 0]
    elif not mult_alp: kappa = kappa[:, 0]
    elif not mult_vw:  kappa = kappa[0, :]

    return kappa

######################### COMPUTING DIAGNOSTIC PROFILES #########################

def w_to_lam(xis, ws, vw, alphan):

    '''
    Function that computes the energy density perturbations
    given the 1d profile of enthalpy using the bag equation
    of state.

    Reference is Appendix A of RoperPol:2025a

    Arguments:
        xis    -- array of xi
        ws     -- array of 1d enthalpies
        vw     -- wall velocity
        alphan -- value of the nucleation alpha

    Returns:
        lam    -- array of energy density perturbations
    '''

    lam        = 3/4*(ws - 1)
    inds       = np.where(xis < vw)[0]
    lam[inds] -= 3/4*alphan

    return lam

########## COMPUTING FUNCTIONS RELEVANT FOR VELOCITY SPECTRAL DENSITY ##########

'''
Main reference is: RoperPol:2025a
Other used references are: Hindmarsh:2016lnk, Hindmarsh:2019phv, RoperPol:2023dzg
'''

#### f' and l functions

def fp_z(xi, vs, z, lz=False, ls=[], multi=True, quiet=False):

    '''
    Function that computes the functions f'(z) and l(z)
    that describe the Fourier transform of the velocity field
    and the energy density perturbations for each expanding
    bubble.

    These functions provide the initial conditions used to
    compute the kinetic spectrum in the sound-wave regime
    according to the Sound-Shell Model.

    Arguments:
        xi    -- array of xi
        vs    -- array of 1d velocities
        z     -- array of z = k (t - t_n) where f' and l functions
                 are to be computed
        lz    -- option to compute l(z) (default is False)
        ls    -- array of energy density perturbations (used if lz
                 is True)
        multi -- option to use an array of wall velocities
        quiet -- option to avoid information printing
                 (default is False)
    '''

    xi_ij, z_ij = np.meshgrid(xi[1:], z, indexing='ij')
    zxi_ij      = z_ij*xi_ij
    j1_z        = np.sin(zxi_ij)/zxi_ij**2 - np.cos(zxi_ij)/zxi_ij
    # avoid division by zero
    j1_z[np.where(zxi_ij == 0)] = 0

    if lz:
        if ls == []:
            print('if lz is chosen you need to provide a l(xi) profile')
            lz = False
        j0_z = np.sin(zxi_ij)/zxi_ij
        j0_z[np.where(zxi_ij == 0)] = 1

    if multi:
        Nvws = np.shape(vs)[0]
        fpzs = np.zeros((Nvws, len(z)))
        if lz: lzs = np.zeros((Nvws, len(z)))

        for i in range(0, Nvws):
            v_ij, z_ij = np.meshgrid(vs[i, 1:], z, indexing='ij')
            try:    fpzs[i, :] = -4*np.pi*np.trapezoid(j1_z*xi_ij**2*v_ij, xi[1:], axis=0)
            except: fpzs[i, :] = -4*np.pi*np.trapz(j1_z*xi_ij**2*v_ij, xi[1:], axis=0)
            if lz:
                l_ij, z_ij = np.meshgrid(ls[i, 1:], z, indexing='ij')
                try:    lzs[i, :]  = 4*np.pi*np.trapezoid(j0_z*xi_ij**2*l_ij, xi[1:], axis=0)
                except: lzs[i, :]  = 4*np.pi*np.trapz(j0_z*xi_ij**2*l_ij, xi[1:], axis=0)

            if not quiet: print('vw ', i + 1, '/', Nvws, ' computed')

    else:

        v_ij, z_ij = np.meshgrid(vs[1:], z, indexing='ij')
        try:    fpzs = -4*np.pi*np.trapezoid(j1_z*xi_ij**2*v_ij, xi[1:], axis=0)
        except: fpzs = -4*np.pi*np.trapz(j1_z*xi_ij**2*v_ij, xi[1:], axis=0)
        if lz:
            l_ij, z_ij = np.meshgrid(ls[1:], z, indexing='ij')
            try:    lzs = 4*np.pi*np.trapezoid(j0_z*xi_ij**2*l_ij, xi[1:], axis=0)
            except: lzs = 4*np.pi*np.trapz(j0_z*xi_ij**2*l_ij, xi[1:], axis=0)

    if lz:
        return fpzs, lzs
    else:
        return fpzs

def Rstar_beta(vws=1., cs2=cs2_ref, corr=True):

    """
    Function that computes the ratio of the mean-bubble separation Rstar
    to the inverse nucleation rate parameter beta.
    """

    Rbeta = (8*np.pi)**(1/3)*vws
    if corr:
        cs = np.sqrt(cs2)
        Rbeta = Rbeta*np.maximum(1., cs/vws)

    return Rbeta



"""
cosmology.py is a Python routine that contains functions relevant
for cosmological calculations, including a solver of Friedmann equations.

Adapted from the original cosmology in GW_turbulence
(https://github.com/AlbertoRoper/GW_turbulence),
created in Nov. 2022

Currently part of the cosmoGW code:

https://github.com/cosmoGW/cosmoGW/
https://github.com/cosmoGW/cosmoGW/blob/main/src/cosmoGW/cosmology.py

Author: Alberto Roper Pol
Created: 27/11/2022 (GW_turbulence)
Updated: 13/03/2025 (release cosmoGW 1.0: https://pypi.org/project/cosmoGW)

Main references are:

RoperPol:2018sap - A. Roper Pol, A. Brandenburg, T. Kahniashvili,
A. Kosowsky, S. Mandal, "The timestep constraint in solving the
gravitational wave equations sourced by hydromagnetic turbulence,"
Geophys. Astrophys. Fluid Dynamics 114, 1, 130 (2020),
arXiv:1807.05479.

RoperPol:2019wvy - A. Roper Pol, S. Mandal, A. Brandenburg,
T. Kahniashvili, A. Kosowsky, "Numerical simulations of gravitational
waves from early-universe turbulence," Phys. Rev. D 102, 083512 (2020),
arXiv:1903.08585.

RoperPol:2022iel - A. Roper Pol, C. Caprini, A. Neronov, D. Semikoz,
"The gravitational wave signal from primordial magnetic fields in the
Pulsar Timing Array frequency band," Phys. Rev. D 105, 123502 (2022),
arXiv:2201.05630.

He:2022qcs       - Y. He, A. Roper Pol, A. Brandenburg,
"Modified propagation of gravitational waves from the early
radiation era," JCAP 06 (2023), 025, arXiv:2212.06082.

Planck:2018vyg   - [Planck], "Planck 2018 results. VI.
Cosmological parameters," Astron.Astrophys. 641 (2020) A6,
Astron.Astrophys. 652 (2021) C4 (erratum),
arXiv:1807.06209.

Rasanen - Lecture notes on Cosmology by Syksy Räsänen
(https://www.mv.helsinki.fi/home/syrasane/cosmo/)

"""

import astropy.constants as const
import astropy.units as u
import pandas as pd
import numpy  as np
from cosmoGW.cosmoGW import HOME

# reference values and constants
Tref     = 100*u.GeV  # EWPT
gref     = 100        # EWPT
Neff_ref = 3          # reference Neff = 3

# values at present time from Planck:2018vyg
T0K      = 2.72548*u.K
H0_ref   = 100*u.km/u.s/u.Mpc
H0_ref   = H0_ref.to(u.Hz)
OmL0_ref = 0.6841
OmM0_ref = 1. - OmL0_ref
h0_ref   = 0.6732

######################### Values at present time #########################

def values_0(h0=1., neut=False, Neff=Neff_ref, ret_rad=False):

    """
    Function that returns the values of cosmological parameters at present time.

    Arguments:
        h0      -- parameterizes the uncertainties (Hubble tension)
                   on the value of the Hubble rate (default 1)
        neut    -- option to add neutrinos in the calculation of today's
                   relativistic dofs
        Neff    -- effective number of neutrino species (default is 3)
        ret_rad -- option to return radiation energy density at present time

    Returns:
        g0      -- 2 relativistic degrees of freedom (assumes massive neutrinos
                   unless ret_rad=True)
        g0s     -- ~3.91 entropic/adiabatic degrees of freedom (including neutrinos)
        T0      -- temperature 2.72548 K, returned in energy units (MeV)
        H0      -- Hubble rate at the present time H0 = 100 h0 km/s/Mpc in frequency
                   units (Hz)
        rho_rad0, Om_rad0 -- radiation energy density (and ratio to critial density)
                             at present time

    Reference: Rasanen.
    """

    # relativistic dofs from photons
    g0  = 2
    # contribution from neutrinos if neut is True
    if neut: g0 = 2*(1 + Neff*7/8*(4/11)**(4/3))
    # adiabatic dofs
    g0s = 2*(1 + Neff*7/8*4/11)
    T0  = (T0K*const.k_B).to(u.MeV)
    H0  = H0_ref*h0

    if ret_rad:
        rho_rad0 = rho_radiation(g=g0, T=T0)
        rho_c    = rho_critical(H=H0)
        Om_rad0  = rho_rad0/rho_c
        return g0, g0s, T0, H0, rho_rad0, Om_rad0
    else: return g0, g0s, T0, H0

########################  Radiation-dominated era ########################

def Hs_fact():

    """
    Function that computes the factor used in the calculation of the Hubble
    parameter at the time of generation (within the radiation-dominated era).
    The factor is given in units of Hz/MeV^2, such that the actual Hubble rate
    is given after multiplying fact*sqrt(g)*T^2, being g the number of dof and
    T the temperature scale (in MeV).

    Reference: RoperPol:2022iel, eq. 29
    """

    fact = np.sqrt(4*np.pi**3*const.G/45/const.c**5/const.hbar**3)
    fact = fact.to(u.Hz/u.MeV**2)

    return fact

def check_temperature_MeV(T, func=''):

    try:
        T = T.to(u.MeV)
    except:
        print('The input temperature in ', func,
              ' needs to be given in energy units using astropy.units \n'
              ' setting T to default value T =', Tref)
        T = Tref.to(u.MeV)

    return T

def Hs_val(g=gref, T=Tref):

    """
    Function that computes the Hubble parameter at the time of generation
    (within the radiation-dominated era).

    Arguments:
        g  -- number of relativistic degrees of freedom (dof) at the
              time of generation (default is 100)
        T  -- temperature scale at the time of generation in energy units
              (convertible to MeV) (default is 100 GeV)

    Returns:
        Hs -- Hubble rate in frequency units (Hz)

    Reference: RoperPol:2022iel, eq. 29
    """

    Hs_f = Hs_fact()
    T    = check_temperature_MeV(T, func='Hs_val')
    Hs   = Hs_f*T**2*np.sqrt(g)

    return Hs

def as_fact(Neff=Neff_ref):

    """
    Function that computes the factor used in the calculation of the ratio
    between the scale factor at the time of generation and the present
    time assuming adiabatic expansion of the universe.

    The factor is in units of MeV and the ratio is obtained by multiplying
    fact*g^(-1/3)/T, being g the number of dof and T the temperature scale
    (in MeV).

    Arguments:
        Neff -- effective number of neutrino species (default is 3)

    Returns:
        fact -- ratio to present-day scale factor, T0 g0s^(1/3)

    Reference: RoperPol:2022iel, eq. 28
    """

    _, g0s, T0, _ = values_0(Neff=Neff)
    fact = T0*g0s**(1/3)

    return fact

def as_a0_rat(g=gref, T=Tref, Neff=Neff_ref):

    """
    Function that computes the ratio between the scale factor at the time
    of generation (a*) and the present time (a_0) assuming adiabatic
    expansion of the universe.

    Arguments:
        g     -- number of relativistic degrees of freedom (dof) at the
                 time of generation (default is 100)
        T     -- temperature scale at the time of generation in energy units
                 (convertible to MeV) (default is 100 GeV)
        Neff  -- effective number of neutrino species (default is 3)

    Returns:
        as_a0 -- ratio of scale factors (a*/a0)

    Reference: RoperPol:2022iel, eq. 28
    """

    as_f  = as_fact(Neff=Neff)
    T     = check_temperature_MeV(T, func='as_a0_rat')
    as_a0 = as_f*g**(-1/3)/T

    return as_a0

############################ COSMOLOGY CALCULATIONS ############################

def rho_radiation(g=gref, T=Tref):

    """
    Function that computes the radiation energy density at different
    epochs of the universe.

    Arguments:
        g -- number of relativistic degrees of freedom (dof) at
             the time of generation (default is 100)
        T -- temperature scale at the time of generation in energy units
             (default is 100 GeV)

    Returns:
        rho_rad -- energy density in GeV/m^3 units

    Reference: RoperPol:2019wvy, eq. 3
    """

    T       = check_temperature_MeV(T, func='rho_radiation')
    rho_rad = np.pi**2/30*g*T**4/(const.hbar*const.c)**3
    rho_rad = rho_rad.to(u.GeV/u.m**3)

    return rho_rad

def check_Hubble_Hz(H, func=''):

    try:
        H = H.to(u.Hz)
    except:
        print('The input Hubble rate in ', func,
              ' needs to be given in frequency units using astropy.units \n'
              ' setting H to default value H =', H0_ref)
        H = H0_ref

    return H

def rho_critical(H=H0_ref):

    """
    Function that computes the critical energy density at different
    epochs of the universe.

    Arguments:
        H -- Hubble rate in units of frequency

    Returns:
        rho_c -- energy density in GeV/m^3 units

    Reference: RoperPol:2019wvy, eq. 3
    """

    H     = check_Hubble_Hz(H, func='rho_critical')
    rho_c = 3*H**2*const.c**2/8/np.pi/const.G
    rho_c = rho_c.to(u.GeV/u.m**3)

    return rho_c

def thermal_g(dir0='', T=Tref, s=0, file=True, Neff=Neff_ref):

    """
    Returns the relativistic dof g_* as a function of T_* according
    to the thermal history of the Universe.
    Note that for T < 0.5 MeV, after neutrino decoupling, entropic
    and relativistic g are approximately equal.

    Arguments:
        dir0 -- directory where the file of dof is stored
                ('resources/cosmology/' directory by default)
        T    -- temperature scale at the time of generation in energy units
                (default is 100 GeV)
        s    -- option to return adiabatic (s=1) dof instead of relativistic
                (default 0)
        file -- option to read g_* or gS from a file with numerical values,
                based on numerical calculations (default True)
                It reads the file 'dir0/cosmology/T_gs.csv' and should be a
                pandas file with columns: ['T [GeV]', 'g_*', 'gS']
        Neff -- effective number of neutrino species (default is 3)

    Returns:
        g    -- relativistic or adiabatic degrees of freedom

    Reference: Rasanen.
    """

    # to make sure that the values at present time correspond to those at the
    # end of RD we take relativistic neutrinos
    g0, g0s, _, _ = values_0(neut=True, Neff=Neff)

    # read the file if it exists
    if file:
        try:
            dirr = dir0 + 'T_gs.csv'
            if dir0 == '':
                dirr = HOME + '/resources/cosmology/T_gs.csv'
            df = pd.read_csv(dirr)
            Ts = np.array(df['T [GeV]'])
            if s == 0: gs = np.array(df['g_*'])
            if s == 1: gs = np.array(df['gS'])
            T  = T.to(u.GeV)    # values of T from file are in GeV
            g  = np.interp(T.value, np.sort(Ts), np.sort(gs))
            T  = check_temperature_MeV(T, func='thermal_g')
            # check if T is a single value or an array
            if not isinstance(T.value, (list, tuple, np.ndarray)):
                if T.value < 0.1:
                    if s == 0: g = g0
                    if s == 1: g = g0s
            else:
                if s == 0: g[T.value < 0.1] = g0
                if s == 1: g[T.value < 0.1] = g0s

        except:
            file = False
            print('thermal_g reads the file %sT_gs.csv, which does not exist! \n'%dir0)
            print('using piecewise approximated function to approximate g')

    # if a file is not given, one can use an approximate piecewise function
    if not file:
        T = check_temperature_MeV(T, func='thermal_g')
        T = T.value
        # check if T is an array or a single value
        if not isinstance(T, (list, tuple, np.ndarray)):
            len_T = 1; T = [T]
        else: len_T = len(T)
        g = np.zeros(len_T)
        for i in range(0, len_T):
            # Check value of T in MeV and assign g_*
            if T[i] < 0.1:
                if s == 0: g[i] = g0
                if s == 1: g[i] = g0s
            elif T[i] < 0.5: g[i] = 7.25
            elif T[i] <= 100: g[i] = 10.75
            elif T[i] <= 150: g[i] = 17.25
            elif T[i] < 1e3: g[i] = 61.75
            elif T[i] < 4e3: g[i] = 75.75
            elif T[i] < 8e4: g[i] = 86.25
            elif T[i] < 1.7e5: g[i] = 96.25
            else: g[i] = 106.75
        if len_T == 1: g = g[0]

    return g

############################### FRIEDMANN EQUATIONS ###############################

"""
The code has been developed and used for the results of
He:2022qcs (see appendix A).

Friedmann solver included to GW_turbulence in June 2022,
adapted and added to cosmoGW for v 1.1.0 release in March 2025.

"""

def RD_dofs(dir0='', Neff=Neff_ref, scalef=False):

    """
    Function that computes the degrees of freedom (relativistic and adiabatic)
    during the RD era.

    Arguments:
        dir0 -- directory where the file of dof is stored
                ('resources/cosmology/' directory by default)'
        Neff -- effective number of neutrino species (default is 3)

    Returns:
        T      -- array of temperatures within RD era
        as_T   -- array of scale factors
        gs, gS -- array of relativistic and adiabatic dofs

    Calls functions 'values_0' and 'thermal_g'; see references therein.
    """

    #### compute the relativistic degrees of freedom as a function
    #### of T during the RD epoch from a stored file using the function
    #### thermal_g

    # array of temperatures during RD era
    T  = np.logspace(-4, 8, 1000)*u.MeV
    gS = thermal_g(T=T, s=1, file=True, dir0=dir0)
    gs = thermal_g(T=T, s=0, file=True, dir0=dir0)

    g0, g0s, T0, _ = values_0(neut=True, Neff=Neff)

    ### convert temperature and degrees of freedom into an array
    ### of scale factors using adiabatic expansion of the universe
    as_T = (T0.to(u.MeV)/T)*(g0s/gs)**(1/3)

    return T, as_T, gs, gS

def dofs_vs_a(a, dir0='', Neff=Neff_ref):

    """
    Function that computes the degrees of freedom (relativistic and adiabatic)
    for an array of scale factors.

    Arguments:
        a    -- array of scale factors
        dir0 -- directory where the file of dof is stored ('/cosmology/' directory by default)
        Neff -- effective number of neutrino species (default is 3)

    Returns:
        gs, gS -- array of relativistic and adiabatic dofs

    Calls functions 'values_0' and 'thermal_g'; see references therein.
    """

    #### Read the dofs during RD era
    T, as_T, gs, gS = RD_dofs(dir0=dir0, Neff=Neff)

    #### convert arrays of degrees of freedom and interpolate to original array of scale factors a
    #### (needs to be sorted first for interp to work)
    inds = np.argsort(as_T)
    as2  = as_T[inds]
    gs   = gs[inds]
    gS   = gS[inds]
    gs   = np.interp(a, as2, gs)
    gS   = np.interp(a, as2, gS)

    return gs, gS

def Hs_from_a(a, dir0='', Neff=Neff_ref):

    """
    Function that computes the Hubble rate H_* during the RD era given only the scale
    factor and assuming adiabatic expansion

        a^3 g_S T^3 = constant

    Arguments:
        a    -- array of scale factors
        dir0 -- directory where the file of dof is stored ('/cosmology/' directory by default)
        Neff -- effective number of neutrino species (default is 3)

    Returns:
        Hs   -- Hubble rate during RD era

    Calls functions 'values_0' and 'thermal_g'; see references therein.
    """

    # compute the dofs
    gs, gS = dofs_vs_a(a, Neff=Neff, dir0=dir0)
    # compute the temperature scale from adiabatic expansions
    g0, g0s, T0, _ = values_0(neut=True, Neff=Neff)
    T = T0.to(u.MeV)/a*(g0s/gs)**(1/3)
    # get the Hubble rate
    Hs = Hs_val(T=T, g=gs)

    return Hs

def Omega_rad_dof(a, dir0='', Neff=Neff_ref):

    """
    Function that computes the factor that takes into account the radiation
    energy density ratio variation due to the inclusion of varying dofs during
    the RD era.

    Arguments:
        a    -- array of scale factors
        dir0 -- directory where the file of dof is stored ('/cosmology/' directory by default)
        Neff -- effective number of neutrino species (default is 3)

    Returns:
        Om_rat_dof -- ratio of radiation energy density due to accounting for
                      T depending dofs

    Reference: He:2022qcs, eq.A.2
    """

    gs, gS = dofs_vs_a(a, dir0=dir0, Neff=Neff)
    g0, g0s, T0, H0 = values_0(neut=True, Neff=Neff)
    Om_rat_dof = (gs/g0)*(gS/g0s)**(-4/3)

    #### the numerical results for the factor (gs/g0)*(gS/g0s)**(-4/3)
    #### present some numerical inacuracy, so it requires some smoothing
    amax, Emax = sp.local_max(a[a>1e-10], Om_rat_dof[a>1e-10])
    Emax[Emax > 1] = 1.
    aa = np.append(a[a<=1e-10], amax)
    Om_rat_dof = np.append(Om_rat_dof[a<=1e-10], Emax)
    Om_rat_dof = np.interp(a, aa, Om_rat_dof)

    return Om_rat_dof

def Omega_vs_a(a, dir0='', a0=1, h0=h0_ref, OmL0=OmL0_ref, dofs=True, Neff=Neff_ref):

    """
    Function that computes the energy density ratio to present-time critical energy
    density as a function of the scale factor a, for a universe composed by matter,
    radiation, and dark energy:

    Omega (a) = OmR0 x a^(-4) + OmM0 x a^(-3) + OmL0

    Arguments:
        a    -- array of scale factors
        dir0 -- directory where the file of dof is stored ('/cosmology/' directory by default)
        a0   -- reference value of the scale factor at present time (default is 1)
        h0   -- present-time value of the Hubble rate H0 = h0 x 100 km/s/Mpc
                (default is 67.32 km/s/Mpc based on CMB observations)
        OmL0 -- present-time content of dark energy
        dofs -- option to compensate the rad energy density using dofs during RD era
        dir0 -- directory where file for dofs during RD era is stored

    Returns:
        Om_tot  -- total energy density (normalized to present-time critical)
        Om_rad  -- radiation energy density (normalized)
        Om_matt -- matter energy density (normalized)

    Reference: He:2022qcs, eq.A.2
    """

    ########## compute Om_rad0 ##########
    g0, g0s, T0, H0, rho_rad0, OmR0 = values_0(h0=h0, neut=True,
                                               Neff=Neff, ret_rad=True)

    #### compute dofs during RD era if dofs is chosen
    #### to compensate the radiation energy density dependence on dofs
    #### during RD era

    if dofs:
        Om_rat_dof=Omega_rad_dof(a, dir0=dir0, Neff=Neff)
        OmR0 = OmR0*Om_rat_dof

    ##### compute other contributions Om_Lam0 and Om_mat0, and total Om_tot
    OmM0   = 1 - OmL0 - OmR0
    Om_mat = (a/a0)**(-3)*OmM0
    Om_rad = (a/a0)**(-4)*OmR0
    Om_tot = OmL0 + Om_rad + Om_mat

    return Om_tot, Om_rad, Om_mat

def friedmann(a, dir0='', a0=1, h0=h0_ref, OmL0=OmL0_ref, dofs=True, Neff=Neff_ref):

    """
    Function that uses Friedmann equations to compute the eos (w) and the time derivatives
    of the scale factor for the radiation energy density described in Omega_vs_a().

    Arguments:
        a    -- array of scale factors
        dir0 -- directory where the file of dof is stored ('/cosmology/' directory by default)
        a0   -- reference value of the scale factor at present time (default is 1)
        h0   -- present-time value of the Hubble rate H0 = h0 x 100 km/s/Mpc
                (default is 67.32 km/s/Mpc based on CMB observations)
        OmL0 -- present-time content of dark energy
        dofs -- option to compensate the rad energy density using dofs during RD era
        dir0 -- directory where file for dofs during RD era is stored

    Returns:
        w   -- equation of state (p = w rho)
        ad  -- cosmic time derivative of the scale factor
        add -- second cosmic time derivative of the scale factor
        ap  -- conformal time derivative of the scale factor
        app -- second conformal time derivative of the scale factor

    Reference: He:2022qcs, eqs. A.5-A.6
    """

    g0, g0s, T0, H0   = values_0(h0=h0, neut=True, Neff=Neff)
    Om_tot, Om_rad, _ = Omega_vs_a(a, a0=a0, h0=h0, OmL0=OmL0, dofs=dofs,
                                   dir0=dir0, Neff=Neff)

    # equation of state
    w   = (1/3*Om_rad - OmL0)/Om_tot
    # time derivatives (cosmic time)
    add = -.5*Om_tot*H0**2*(a/a0)*(1 + 3*w)
    ad  = (a/a0)*np.sqrt(Om_tot)*H0
    # time derivatives (conformal time)
    app = .5*a**3*Om_tot*(1 - 3*w)*H0**2
    ap  = ad*a

    return w, ad, add, ap, app

def friedmann_solver(a, dir0='', a0=1., h0=h0_ref, OmL0=OmL0_ref,
                     dofs=True, Neff=Neff_ref,
                     return_all=False, save=True, nm_fl=''):

    """
    Function that uses Friedmann equations and solve them numerically to obtain
    the evolution of a(eta) and a(t) using the Omega distribution obtained from
    the function Omega_vs_a.

    A tutorial is available under cosmology/cosmology.ipynb

    Arguments:
        a    -- array of scale factors
        dir0 -- directory where the file of dof is stored ('/cosmology/' directory by default)
        a0   -- reference value of the scale factor at present time (default is 1)
        h0   -- present-time value of the Hubble rate H0 = h0 x 100 km/s/Mpc
                (default is 67.32 km/s/Mpc based on CMB observations)
        OmL0 -- present-time content of dark energy (default is 0.6841)
        dofs -- option to compensate the rad energy density using dofs during RD era
        Neff -- effective number of neutrino species (default is 3)
        return_all -- option to return all variables used in the Friedmann solver
        save -- option to save the solutions in the file
                'friedmann/solution#nm_fl.csv' where #nm_fl is the inpute file name
                The input variables used in the solver are stored in
                'friedmann/README#nm_fl.csv'

    Reference: He:2022qcs, appendix A.
    """

    Om_tot, Om_rad, Om_mat = Omega_vs_a(a, a0=a0, h0=h0, OmL0=OmL0, dir0=dir0,
                                         dofs=dofs, Neff=Neff)

    #### numerically compute the arrays of cosmic and conformal times
    #### from the energy density ratio
    difft   = np.zeros(len(a))
    diffeta = np.zeros(len(a))
    if a[0] > 1e-18:
        print('minimum a given is ', a[0])
        print('note that the solver assumes that a[0] corresponds to',
              'times t = eta = 0, so make sure to use an array of a with',
              'small enough values (a[0] ~ 1e-20)')
    difft[0]   = 0
    diffeta[0] = 0

    print('Entering Friedmann solver')

    for i in range(1, len(a)):
        aas   = np.logspace(np.log10(a[0]), np.log10(a[i]), 10000)
        Om_as = np.interp(aas, a, Om_tot)
        fft   = 1/aas/np.sqrt(Om_as)
        ffeta = fft/aas
        try:
            difft[i]   = np.trapezoid(fft, aas)
            diffeta[i] = np.trapezoid(ffeta, aas)
        except:
            difft[i]   = np.trapz(fft, aas)
            diffeta[i] = np.trapz(ffeta, aas)

    g0, g0s, T0, H0 = values_0(h0=h0)
    t   = difft/H0
    eta = diffeta/H0

    # Using Friedmann equations, we compute the eos (w), and the time derivatives of the
    # scale factor
    w, ad, add, ap, app = friedmann(a, dir0=dir0, h0=h0, OmL0=OmL0, dofs=dofs, Neff=Neff)

    if save:
        df = pd.DataFrame({'a' : a, 't' : t, 'eta': eta,
                           'ap/a': ap/a, 'app/a' : app/a})
        df.to_csv('friedmann/solution' + nm_fl + '.csv')
        with open('friedmann/README' + nm_fl + '.txt', 'w') as f:
            f.write("The file solution" + nm_fl + ".csv contains the ")
            f.write("solutions from the Friedmann solver using the parameters:\n")
            f.write("a0 = %.4e, h0 = %.4f, OmL0 = %.4f, Neff = %.4f \n"%(a0, h0, OmL0, Neff))
            f.write("The solution contains 'a', 't', 'eta', 'ap/a' and 'app/a'")
        print('The results are saved in the file friedmann/solution' + nm_fl + '.csv')
        print('The input parameters used are stored in friedmann/README' + nm_fl + '.txt')

    print('Leaving Friedmann solver')

    if return_all: return t, eta, Om_tot, Om_rad, Om_mat, w, ad, add, ap, app
    else: return t, eta

def normalized_variables(a, eta, ap_a, app_a, dir0='', T=Tref, h0=h0_ref):

    """
    Function that computes the normalized a, eta, HH, a'' for a given specific
    initial time of GW generation, which are required to be used in the Pencil Code.

    A tutorial is available under cosmology/cosmology_PC.ipynb

    Arguments:
        a     -- scale factors, normalized to present-time a_0 = 1
        eta   -- conformal times, normalized to present-time a_0 = 1
        ap_a  -- conformal Hubble time a'/a, normalized to present-time a_0 = 1
        app_a -- a''/a, normalized to present-time a_0 = 1
        dir0  -- directory where the file of dof is stored ('/cosmology/' directory by default)
        T     -- temperature scale at the time of generation in energy units
                 (default is 100 GeV)
        h0    -- present-time value of the Hubble rate H0 = h0 x 100 km/s/Mpc
                 (default is 67.32 km/s/Mpc based on CMB observations)

    Returns:
        a_n       -- normalized scale factor a/a_*
        eta_n     -- normalized conformal time eta/eta_*
        HH_n      -- normalized conformal Hubble rate H/H_*
        app_a_n   -- normalized second conformal time derivative of a, (a''/a)/H_*^2
        Omega     -- ratio of total energy to present-time critical energy denstiy
        w         -- equation of state
        eta_n_0   -- normalized conformal present time
        aEQ_n     -- normalized equipartition scale factor
        aL_n      -- normalized dark energy domination scale factor
        a_acc_n   -- normalized scale factor when acceleration starts
        eta_n_EQ  -- normalized conformal time at equipartition
        eta_n_L   -- normalized conformal time at dark energy domination
        eta_n_acc -- normalized conformal time when acceleration starts

    Reference: He:2022qcs, appendix A.

    The normalization follows that from RoperPol:2018sap.
    """

    g   = thermal_g(T=T, s=0, dir0=dir0)
    gS  = thermal_g(T=T, s=1, dir0=dir0)
    ast = as_a0_rat(T=T, g=gS)
    Hs  = Hs_val(T=T, g=g)

    # normalized scale rates
    a_n = a/ast
    a0  = 1/ast

    ## Find value of eta at the initial time of generation
    eta_ast = np.interp(ast, a, eta)
    # normalized conformal times
    eta_n   = eta/eta_ast
    eta_n_0 = np.interp(a0, a_n, eta_n)

    # normalized conformal Hubble rate
    HH_n = ap_a/Hs/ast

    # normalized acceleration
    app_a_n = app_a/Hs**2/ast**2

    ### we can recover the values of Omega and w
    H0    = h0*H0_ref
    Omega = (HH_n.value*Hs)**2/a_n**2/H0**2
    w     = 1/3*(1 - app_a_n*2/HH_n**2)

    # compute aEQ, aL, and a_acc
    inds      = np.argsort(w)
    aEQ_n     = np.interp(1/6, w[inds], a_n[inds])
    aL_n      = np.interp(-.5, w[inds], a_n[inds])
    a_acc_n   = np.interp(-1/3, w[inds], a_n[inds])
    eta_n_EQ  = np.interp(aEQ_n, a_n, eta_n)
    eta_n_L   = np.interp(aL_n, a_n, eta_n)
    eta_n_acc = np.interp(a_acc_n, a_n, eta_n)

    return a_n, eta_n, HH_n, app_a_n, Omega, w, eta_n_0, aEQ_n, \
           aL_n, a_acc_n, eta_n_EQ, eta_n_L, eta_n_acc

def ratio_app_a_n_factor(a, dir0='', a0=1, h0=h0_ref, OmL0=OmL0_ref,
                         dofs=True, Neff=Neff_ref):

    """
    Function that computes the ratio of a''/a (normalized) to conformal Hubble rate H (normalized)
    times a_*/a_0 using an approximation valid during the RD era.

    Reference: He:2022qcs, eq. 3.11

    Arguments:
        a    -- array of scale factors
        dir0 -- directory where the file of dof is stored ('/cosmology/' directory by default)
        a0   -- reference value of the scale factor at present time (default is 1)
        h0   -- present-time value of the Hubble rate H0 = h0 x 100 km/s/Mpc
                (default is 67.32 km/s/Mpc based on CMB observations)
        OmL0 -- present-time content of dark energy (default is 0.6841)
        dofs -- option to compensate the rad energy density using dofs during RD era
        Neff -- effective number of neutrino species (default is 3)

    Returns:
        factor -- ratio of a''/a to conformal Hubble rate H times a_*/a_0
    """

    g0, g0s, T0, H0, rho_rad0, OmR0 = values_0(h0=h0, neut=True, Neff=Neff, ret_rad=True)
    OmM0 = 1 - OmL0 - OmR0
    Om_rat_dof = Omega_rad_dof(a, Neff=Neff, dir0=dir0)
    factor = .5*OmM0/OmR0/Om_rat_dof

    return factor

def norm_variables_cut(eta_n, HH_n, a_n, Omega, Omega_mat, eta_n_0, dir0='',
                       T=Tref, OmM0=OmM0_ref, h0=h0_ref):

    """
    Function that cuts the normalized variables between the initial time eta/eta_* = 1
    to present-time.

    Arguments:
        eta_n   -- normalized conformal time eta/eta_*
        HH_n    -- normalized conformal Hubble rate H/H_*
        a_n     -- normalized scale factor a/a_*
        Omega   -- ratio of total energy to present-time critical energy denstiy
        Om_mat  -- matter energy density (normalized)
        eta_n_0 -- normalized conformal present time
        Hs      -- Hubble rate at the initial time
        ast     -- scale factor at the initial time
        dir0    -- directory where the file of dof is stored
                   ('/cosmology/' directory by default)
        T       -- temperature scale at the initial time in energy units
                   (default is 100 GeV)
        OmM0    -- present-time content of matter (default is 0.3159)
        h0      -- present-time value of the Hubble rate H0 = h0 x 100 km/s/Mpc
                   (default is 67.32 km/s/Mpc based on CMB observations)

    Returns: variables given as arrays from the initial time until present time
        eta_nn       -- normalized conformal time
        HH_nn        -- normalized conformal Hubble rate
        a_nn         -- normalized scale factor
        Omega_nn     -- ratio of total energy to present-time critical energy denstiy
        Omega_mat_nn -- matter energy density (normalized)
        app_nn       -- second time derivative of the scale factor
        w_nn         -- equation of state p/rho
    """

    H0 = h0*H0_ref

    # relativistic and adiabatic dofs
    g   = thermal_g(T=T, s=0, dir0=dir0)
    gS  = thermal_g(T=T, s=1, dir0=dir0)
    # scale factor and Hubble rate
    ast = as_a0_rat(T=T, g=gS)
    Hs  = Hs_val(T=T, g=g)

    # indices
    inds   = np.where(eta_n > 1)[0]
    inds2  = np.where(eta_n[inds] < eta_n_0)[0]

    eta_nn = cut_var(eta_n, 1, eta_n_0, inds, inds2)
    HH_n0  = np.interp(1, eta_n, HH_n.value)
    HH_nn  = cut_var(HH_n.value, HH_n0, H0/Hs/ast, inds, inds2)
    a_nn   = cut_var(a_n, 1, 1/ast, inds, inds2)
    Omega_nn = cut_var(Omega, (Hs/H0)**2, 1, inds, inds2)
    Omega_mat_nn = cut_var(Omega_mat, OmM0*ast**(-3), OmM0, inds, inds2)
    Omega_rad_nn = Omega_nn - Omega_mat_nn - (1 - OmM0)
    w_nn   = (1/3*Omega_rad_nn - (1 - OmM0))/Omega_nn
    app_nn = .5*HH_nn**2*(1 - 3*w_nn)

    return eta_nn, HH_nn, a_nn, Omega_nn, Omega_mat_nn, app_nn, w_nn

def cut_var(x, x0, xf, inds, inds2):

    """
    Function that cuts the variable x using the indices inds and inds2
    and adding an initial value x0 and final value xf.

    Used in norm_variables_cut function.
    """

    y = x[inds][inds2]
    y = np.append(x0, y)
    y = np.append(y, xf)

    return y



"""
GW_models.py is a Python routine that contains analytical and
semi-analytical models of cosmological GW backgrounds.

Currently part of the cosmoGW code:

https://github.com/cosmoGW/cosmoGW/
https://github.com/cosmoGW/cosmoGW/blob/main/src/cosmoGW/GW_models.py

Author: Alberto Roper Pol
Created: 29/08/2024
Updated: 04/06/2025 (release cosmoGW 1.0: https://pypi.org/project/cosmoGW)

Other contributors: Antonino Midiri, Simona Procacci, Madeline Salomé

Main references are:

RoperPol:2022iel  - A. Roper Pol, C. Caprini, A. Neronov, D. Semikoz,
"The gravitational wave signal from primordial magnetic fields in the
Pulsar Timing Array frequency band," Phys. Rev. D 105, 123502 (2022),
arXiv:2201.05630

RoperPol:2023bqa  - A. Roper Pol, A. Neronov, C. Caprini, T. Boyer,
D. Semikoz, "LISA and γ-ray telescopes as multi-messenger probes of a
first-order cosmological phase transition," arXiv:2307.10744 (2023)

RoperPol:2023dzg  - A. Roper Pol, S. Procacci, C. Caprini,
"Characterization of the gravitational wave spectrum from sound waves within
the sound shell model," Phys. Rev. D 109, 063531 (2024), arXiv:2308.12943

Hindmarsh:2019phv - M. Hindmarsh, M. Hijazi, "Gravitational waves from first order
cosmological phase transitions in the Sound Shell Model,"
JCAP 12 (2019) 062, arXiv:1909.10040

Caprini:2024gyk   - A. Roper Pol, I. Stomberg, C. Caprini, R. Jinno,
T. Konstandin, H. Rubira, "Gravitational waves from first-order
phase transitions: from weak to strong," JHEP, arxiv:2409.03651

RoperPol:2025b    - A. Roper Pol, A. Midiri, M. Salomé, C. Caprini,
"Modeling the gravitational wave spectrum from slowly decaying sources in the
early Universe: constant-in-time and coherent-decay models," in preparation

RoperPol:2025a    - A. Roper Pol, S. Procacci, A. S. Midiri,
C. Caprini, "Irrotational fluid perturbations from first-order phase
transitions," in preparation
"""

import numpy as np
import matplotlib.pyplot as plt
import cosmoGW.hydro_bubbles as hb
import cosmoGW.GW_analytical as an

# reference values
cs2_ref    = hb.cs2_ref  # speed of sound squared
Oms_ref    = 0.1         # reference source amplitude
                         # (fraction to radiation energy)
lf_ref     = 0.01        # reference length scale of the source
                         # (normalized by the Hubble radius)
N_turb     = 2           # ratio between the effective time duration of
                         # the source and the eddy turnover time,
                         # based on the simulations of RoperPol:2022iel,
                         # used in RoperPol:2023bqa
Nk_ref     = 1000        # reference number of wave number discretization
Nkconv_ref = 1000        # reference number of wave number discretization
                         # for convolution calculations
Np_ref     = 3000        # reference number of wave number discretization
                         # for convolution calculations
NTT_ref    = 5000        # reference number of lifetimes discretization
dt0_ref    = 11          # dt0 = 11 is a numerical parameter of the fit
                         # provided in Caprini:2024gyk

'''
    RoperPol:2022iel/RoperPol:2023dzg and RoperPol:2025b/RoperPol:2025b
    consider spectral functions defined such that the average squared
    field corresponds to

    <v^2> ~ 2 E* k* int zeta(K) dK,     in RoperPol:2022iel
    <v^2> ~ 2 E*    int zeta(K) dlnK,   in RoperPol:2025b

    The first convention can be chosen in the following functions if
    dlogK is set to False, while the second one is assumed when dlogK
    is True
'''

################ COMPUTING THE SPECTRUM OF THE STRESSES  ###############
######################## FOR DIFFERENT SOURCES #########################

def Integ(p, tildep, z, k=0., tp='vort', hel=False):

    '''
    Integrand of the integral over p and z (or tilde p) that is
    used to compute the anisotropic stresses in EPi_correlators
    '''

    Integ = 0
    if hel:
        if tp == 'vort': Integ = 1./p/tildep**4*(1 + z**2)*(k - p*z)
        if tp == 'comp': Integ = 2./p/tildep**4*(1 - z**2)*(k - p*z)
    else:
        if tp == 'vort': Integ = .5/p/tildep**3*(1 + z**2)* \
                                     (2 - p**2/tildep**2*(1 - z**2))
        if tp == 'comp': Integ = 2.*p/tildep**5*(1 - z**2)**2
        if tp == 'mix':  Integ = 2.*p/tildep**5*(1 - z**4)
        if tp == 'hel':  Integ = .5/p/tildep**4*z*(k - p*z)

    return Integ

def EPi_correlators(k, a=an.a_ref, b=an.b_ref, alp=an.alp_ref,
                    tp='all', zeta=False, hel=False, norm=True,
                    dlogk=True, model='dbpl', kk=[], EK_p=[]):

    '''
    Routine to compute the spectrum of the projected (anisotropic) or
    unprojected stresses from the two-point correlator of the source
    (e.g. velocity, magnetic or scalar fields) under the assumption that
    the source is Gaussian.

    It computes the vortical, compressional, helical and mixed components
    of the stresses, as well as the compressional and vortical components
    of the helical stresses.

    Main reference is RoperPol:2025a, see also eqs 11 and 20 of
    RoperPol:2022iel for the vortical component and eq. 45 of
    RoperPol:2023dzg for the compressional component

    Arguments:
        k     -- array of wave numbers
        a     -- slope of the spectrum at low wave numbers, k^a
        b     -- slope of the spectrum at high wave numbers, k^(-b)
        alp   -- smoothness of the transition from one power law to the other
        tp    -- type of sourcing field: 'vort', 'comp', 'hel' or 'mix' available
        zeta  -- option to integrate the convolution over z in (-1, 1)
                default option integrates over ptilde in (|p - k|, p + k)
        hel   -- option to compute the helical stresses
        norm  -- option to normalize the spectrum such that its peak is located at
                 kpeak and its maximum value is A
        model -- select the model to be used for the spectrum of the source,
                 current options are 'dbpl' for the smoothed double broken power
                 law model defined in GW_analytical module or 'input' to give
                 a numerical input
        kk, EK_p -- if input is chosen, then numerical array of wave numbers and
                    spectrum of the source need to be provided

    Returns
        pi -- spectrum of the anisotropic stresses of the chosen component
              or of all of them if tp = 'all' is chosen
    '''

    from scipy import integrate

    # define zeta_P times zeta_Ptilde function in funcs based on
    # the inpute model

    if model == 'dbpl':

        A    = (a + b)**(1/alp)
        alp2 = alp*(a + b)
        if norm: c = a; d = b
        else: c = 1.; d = 1.; A = 1.

        # functions following a smoothed broken power law
        def funcs(p, tildep):
            zeta_P      = A*p**a/(d + c*p**alp2)**(1/alp)
            zeta_Ptilde = A*tildep**a/(d + c*tildep**alp2)**(1/alp)
            return zeta_P*zeta_Ptilde

    elif model == 'input':

        if len(kk) == 0 or len(EK_p) == 0:
            print('For using input model provide kk and EK')
            return 0.

        # functions interpolate the input numerical data
        def funcs(p, tildep):
            zeta_P      = np.interp(p, kk, EK_p)
            zeta_Ptilde = np.interp(p, kk, EK_p)
            return zeta_P*zeta_Ptilde

    else:
        print('A model needs to be selected: dpbl or input')
        return 0.

    ## compute all components (vort, comp, mix, hel)

    if tp == 'all':
        if hel: tps = ['vort', 'comp']
        else:   tps = ['vort', 'comp', 'mix', 'hel']
        pis = np.zeros((len(tps), len(k)))
        for j in range(0, len(tps)):
            # integrate over p and z
            if zeta:

                def f(p, z, kp):
                    tildep = np.sqrt(p**2 + kp**2 - 2*p*kp*z)
                    II     = funcs(p, tildep)*Integ(p, tildep, z, k=kp,
                                                    tp=tps[j], hel=hel)
                    if not dlogk: II *= p*tildep
                    return II

                for i in range(0, len(k)):
                    kp = k[i]
                    pis[j, i], _ = \
                        integrate.nquad(f, [[0, np.inf], [-1., 1.]], args=(kp,))

            # integrate over p and ptilde
            else:

                for i in range(0, len(k)):
                    kp = k[i]

                    def f(p, tildep):
                        z  = (p**2 + kp**2 - tildep**2)/2/p/kp
                        II = funcs(p, tildep)*Integ(p, tildep, z, k=kp,
                                   tp=tps[j], hel=hel)*tildep/p/kp
                        if not dlogk: II *= p*tildep
                        return II

                    def bounds_p():       return [0, np.inf]
                    def bounds_tildep(p): return [abs(kp - p), kp + p]

                    pis[j, i], _ = integrate.nquad(f, [bounds_tildep, bounds_p])

        return pis

    ## compute the chosen component in tp
    else:

        pi = np.zeros(len(k))

        # integrate over p and z
        if zeta:

            def f(p, z, kp):
                tildep = np.sqrt(p**2 + kp**2 - 2*p*kp*z)
                II     = funcs(p, tildep)*Integ(p, tildep, z, k=kp,
                               tp=tp, hel=hel)
                if not dlogk: II *= p*tildep
                return II

            for i in range(0, len(k)):
                kp = k[i]
                pi[i], _ = \
                    integrate.nquad(f, [[0, np.inf], [-1., 1.]], args=(kp,))

        # integrate over p and ptilde
        else:
            for i in range(0, len(k)):
                kp = k[i]
                def f(p, tildep):
                    z  = (p**2 + kp**2 - tildep**2)/2/p/kp
                    II = funcs(p, tildep)*Integ(p, tildep, z, k=kp,
                               tp=tp, hel=hel)*tildep/p/kp
                    if not dlogk: II *= p*tildep
                    return II

                def bounds_p(): return [0, np.inf]
                def bounds_tildep(p): return [abs(kp - p), kp + p]

                pi[i], _ = integrate.nquad(f, [bounds_tildep, bounds_p])

        return pi

###########  FUNCTIONS FOR THE CONSTANT-IN-TIME MODEL ###########

def Delta_cit(t, k, tini=1, tfin=1e4, expansion=True):

    """
    Function that computes the value of the function Delta(k, t) used in the
    analytical calculations of the GW spectrum when assuming
    a constant sourcing stress spectrum, i.e., Pi(k, t1, t2) = Pi(k)

    Arguments:
        k         -- array of wave numbers
        t         -- array of times
        tini      -- initial time of the turbulence sourcing (default 1)
        tfin      -- final time of the turbulence sourcing
        expansion -- option to include the expansion of the Universe
                     during radiation domination (default is True)

    Returns:
        Delta     -- function Delta(k, t)
    """

    import scipy.special as spe

    mult_t = isinstance(t, (list, tuple, np.ndarray))
    mult_k = isinstance(k,  (list, tuple, np.ndarray))

    if not mult_t: t = [t]
    if not mult_k: k = [k]

    tij, kij = np.meshgrid(t, k, indexing='ij')
    cost     = np.cos(kij*tij)
    sint     = np.sin(kij*tij)
    tij[np.where(tij>tfin)] = tfin
    if expansion:
        si_t, ci_t       = spe.sici(kij*tij)
        si_tini, ci_tini = spe.sici(kij*tini)
    else:
        ci_t    =  np.sin(kij*tij)/kij
        ci_tini =  np.sin(kij*tini)/kij
        si_t    = -np.cos(kij*tij)/kij
        si_tini = -np.cos(kij*tini)/kij

    aux1 = cost*(ci_t - ci_tini)
    aux2 = sint*(si_t - si_tini)
    D    = aux1 + aux2

    if   not mult_t and not mult_k: D = D[0, 0]
    elif not mult_t: D = D[0, :]
    elif not mult_k: D = D[:, 0]

    return D

def Delta2_cit_aver(k, tini=1, tfin=1e4, expansion=True):

    """
    Function that computes the value of the function D(k, t) used in the
    analytical calculations of the GW energy density spectrum when assuming
    a constant sourcing stress spectrum, i.e., Pi(k, t1, t2) = Pi(k)

    Arguments:
        k         -- array of wave numbers
        tini      -- initial time of the turbulence sourcing (default 1)
        tfin      -- final time of the turbulence sourcing
        expansion -- option to include the expansion of the Universe
                     during radiation domination (default is True)

    Returns:
        D2 -- function D^2(k, t) averaged over t0
    """

    import scipy.special as spe

    if expansion:
        si_t, ci_t       = spe.sici(k*tfin)
        si_tini, ci_tini = spe.sici(k*tini)
    else:
        ci_t    =  np.sin(k*tfin)/k
        ci_tini =  np.sin(k*tini)/k
        si_t    = -np.cos(k*tfin)/k
        si_tini = -np.cos(k*tini)/k
    aux1 = (ci_t - ci_tini)
    aux2 = (si_t - si_tini)
    D2   = .5*(aux1**2 + aux2**2)

    return D2

def TGW_func(s, Oms=Oms_ref, lf=lf_ref, N=N_turb, cs2=cs2_ref,
             expansion=True, tdecay='eddy', tp='magnetic'):

    """
    Function that computes the logarithmic term obtained as
    the envelope of the GW template in the constant-in-time (cit)
    assumption for the unequal time correlator of the turbulent
    stresses.

    It is obtained from integrating the Green's function in the
    constant-in-time model, validated in RoperPol:2022iel
    and described for other sources in RoperPol:2025b

    The function TGW_func determines two regimes, frequencies below
    and above an inverse duration,

        fbr = 1/dtfin,

    as used in RoperPol:2023bqa, EPTA:2023xxk, Caprini:2024hue.
    However, the original reference RoperPol:2022iel used

        kbr = fbr/(2pi) = 1/dtfinm,

    such that dtfinm = dtfin/(2pi), and relates dtfin to
    the decay time (considered the eddy turnover time) as

        dtfinm = N tdecay = N/(vA k*) = N R*/(2pi vA)

    Hence, we will consider

        fbr   = 1/dtfin = 1/(2pi dtfinm) = vA/(N R*)
        dtfin = N R*/vA

    Main references are:

    - RoperPol:2022iel, equation 24
    - RoperPol:2023bqa, equation 15
    - EPTA:2023xxk,     equation 21
    - Caprini:2024hue,  equation 2.17
    - RoperPol:2025b

    Arguments:
        s      -- array of frequencies, normalized by the
                 characteristic scale, s = f R*
        Oms    -- energy density of the source (i.e., 1/2 vrms^2)
        lf     -- characteristic scale of the turbulence as a
                  fraction of the Hubble radius, R* H*
        N      -- relation between the decay time and the effective
                  source duration
        cs2    -- square of the speed of sound (default is 1/3)
        expansion -- option to include the expansion of the Universe
                     during radiation domination (default is True)
        tdecay -- determines the finite duration in the cit model
                  (default is to use eddy turnover time)
        tp     -- type of source (default is 'magnetic', other options are
                  'kinetic' or 'max') used to compute the characteristic
                  velocity in the eddy turnover time

    Returns:
        TGW   -- logarithmic function that characterizes the spectral shape
    """

    mult_Oms = isinstance(Oms, (list, tuple, np.ndarray))
    mult_lf  = isinstance(lf,  (list, tuple, np.ndarray))

    if not mult_Oms: Oms = [Oms]
    if not mult_lf:  lf  = [lf]

    s, Oms, lf = np.meshgrid(s, Oms, lf, indexing='ij')

    # characteristic velocity (for example, Alfven velocity or vrms)
    # see eq. 12 of RoperPol:2023bqa
    if   tp == 'kinetic':  vA = np.sqrt(Oms)
    elif tp == 'magnetic': vA = np.sqrt(2*Oms/(1 + cs2))
    else:                  vA = np.sqrt(max(1, 2/(1 + cs2))*Oms)
    # decay time in units of R*
    if   tdecay == 'eddy': tdec = 1./vA

    # effective duration of the source dtfin/R* is N units of
    # the decay time (see comment above for different choices
    # in the literature)
    dtfin = N*tdec

    TGW = np.zeros_like(dtfin)
    if expansion:
        TGW[s <  1/dtfin] = np.log(1 + dtfin*lf/2/np.pi)**2
        TGW[s >= 1/dtfin] = np.log(1 + lf/2/np.pi/s)**2
    else:
        TGW[s <  1/dtfin] = (dtfin*lf/2/np.pi)**2
        TGW[s >= 1/dtfin] = (lf/2/np.pi/s)**2


    if not mult_Oms and not mult_lf: TGW = TGW[:, 0, 0]
    elif not mult_Oms:               TGW = TGW[:, 0, :]
    elif not mult_lf:                TGW = TGW[:, :, 0]

    return TGW

################ SOUND-SHELL MODEL FOR SOUND WAVES IN PTs ################

'''
Kinetic spectra computed for the sound-shell model from f' and l functions.
f' and l functions need to be previously computed from the self-similar
fluid perturbations induced by expanding bubbles using hydro_bubbles module
'''

def compute_kin_spec_ssm(z, vws, fp, l=[], sp='sum', type_n='exp', cs2=cs2_ref,
                         min_qbeta=-4, max_qbeta=5, Nqbeta=Nk_ref,
                         min_TT=-1, max_TT=3, NTT=NTT_ref, corr=False,
                         dens=True, normbeta=True):

    '''
    Function that computes the kinetic power spectral density assuming
    exponential or simultaneous nucleation.

    Reference: Equations (32)--(37) of RoperPol:2023dzg
    Main reference is RoperPol:2025a

    Arguments:
        z      -- array of values of z
        vws    -- array of wall velocities
        fp     -- function f'(z) computed from the hydro_bubble module using fp_z
        l      -- function lambda(z) computed from the hydro_bubble module using fp_z
                  (using lz = True)
        sp     -- type of function computed for the kinetic spectrum description
        type_n -- type of nucleation hystory (default is exponential 'exp',
                  another option is simultaneous 'sym')
        cs2    -- square of the speed of sound (default 1/3)
        corr   -- option to 'correct' Rstar beta with max(vw, cs) (default is False)
        dens   -- option to return power spectral density (if True, default), or kinetic
                  spectrum (if False)
        normbeta -- normalization of k with beta (default is True).
                    If normbeta is False, k is normalized with Rstar

    Returns:
        qbeta -- wave number, normalized with 1/beta or Rstar
        P     -- power spectral density (power spectrum if dens is False)
    '''

    if sp == 'sum':    A2 = .25*(cs2*l**2 + fp**2)
    if sp == 'only_f': A2 = .5*fp**2
    if sp == 'only_l': A2 = .5*cs2*l**2
    if sp == 'diff':   A2 = .25*(fp**2 - cs2*l**2)
    if sp == 'cross':  A2 = -.5*fp*np.sqrt(cs2)*l

    qbeta = np.logspace(min_qbeta, max_qbeta, Nqbeta)
    TT    = np.logspace(min_TT, max_TT, NTT)
    Pv    = np.zeros((len(vws), len(qbeta)))

    q_ij, TT_ij = np.meshgrid(qbeta, TT, indexing='ij')
    if type_n == 'exp': nu_T = np.exp(-TT_ij)
    if type_n == 'sim': nu_T = .5*np.exp(-TT_ij**3/6)*TT_ij**2

    funcT = np.zeros((len(vws), len(qbeta), len(TT)))

    for i in range(0, len(vws)):
        funcT[i, :, :]   = nu_T*TT_ij**6*np.interp(TT_ij*q_ij, z, A2[i, :])
        try:    Pv[i, :] = np.trapezoid(funcT[i, :, :], TT, axis=1)
        except: Pv[i, :] = np.trapz(funcT[i, :, :], TT, axis=1)

    if dens == False:
        Rstar_beta   = hb.Rstar_beta(vw=vws[i], cs2=cs2, corr=corr)
        for i in range(0, len(vws)):
            pref     = qbeta[i, :]**2/Rstar_beta[i]**4/(2*np.pi**2)
            Pv[i, :] *= pref

    if normbeta == False:
        if not dens: Rstar_beta = hb.Rstar_beta(vw=vws[i], cs2=cs2, corr=corr)
        for i in range(0, len(vws)):
            qbeta[i, :] *= Rstar_beta[i]

    return qbeta, Pv

def OmGW_ssm_HH19(k, EK, Np=Np_ref, Nk=Nkconv_ref, plot=False,
                  cs2=cs2_ref):

    '''
    Function to compute GW spectrum using the approximation
    introduced in the first sound-shell model analysis of Hindmarsh:2019phv
    under the delta assumption for the 'stationary' term
    (see appendix B of RoperPol:2023dzg for details).

    The resulting GW spectrum is

     Omega_GW (k) = (3pi)/(8cs) x Gamma2 x (k/kst)^2 x (Oms/cal A)^2 x TGW x Omm(k)

     where Oms = <v^2> = vrms^2 and cal A = int zeta(K) d ln k
     (using the normalization of RoperPol:2025a for zeta)

    Reference: Appendix B of RoperPol:2023dzg; see eq.(B3)
    '''

    cs    = np.sqrt(cs2)
    kp    = np.logspace(np.log10(k[0]), np.log10(k[-1]), Nk)

    p_inf = kp*(1 - cs)/2/cs
    p_sup = kp*(1 + cs)/2/cs

    Omm = np.zeros(len(kp))
    for i in range(0, len(kp)):

        p      = np.logspace(np.log10(p_inf[i]), np.log10(p_sup[i]), Np)
        ptilde = kp[i]/cs - p
        z      = -kp[i]*(1 - cs2)/2/p/cs2 + 1/cs

        EK_p      = np.interp(p, k, EK)
        EK_ptilde = np.interp(ptilde, k, EK)

        Omm1   = (1 - z**2)**2*p/ptilde**3*EK_p*EK_ptilde
        try:    Omm[i] = np.trapezoid(Omm1, p)
        except: Omm[i] = np.trapz(Omm1, p)

    return kp, Omm

def effective_ET_correlator_stat(k, EK, tfin, Np=Np_ref, Nk=Nkconv_ref,
                                 plot=False, expansion=True, kstar=1.,
                                 extend=False, largek=3, smallk=-3, tini=1,
                                 cs2=cs2_ref, terms='all',
                                 inds_m=[], inds_n=[], corr_Delta_0=False):

    """
    Function that computes the normalized GW spectrum zeta_GW(k)
    from the velocity field spectrum for purely compressional anisotropic stresses,
    assuming Gaussianity, and under the assumption of stationary UETC (e.g.,
    sound waves under the sound-shell model).

    Reference: RoperPol:2023dzg, eq. 93

    Arguments:
        k      -- array of wave numbers
        EK     -- array of values of the kinetic spectrum
        Np     -- number of discretizations in the wave number p to be numerically
              integrated
        Nk     -- number of discretizations of k to be used for the computation of
              the final spectrum
        plot   -- option to plot the interpolated magnetic spectrum for debugging
                purposes (default False)
        extend -- option to extend the array of wave numbers of the resulting
                  Pi spectrum compared to that of the given magnetic spectrum
                  (default False)

    Returns:
        Omm -- GW spectrum normalized (zeta_GW)
        kp  -- final array of wave numbers
    """

    p  = np.logspace(np.log10(k[0]), np.log10(k[-1]), Np)
    kp = np.logspace(np.log10(k[0]), np.log10(k[-1]), Nk)
    if extend:
        Nk = int(Nk/6)
        kp = np.logspace(smallk, np.log10(k[0]), Nk)
        kp = np.append(kp, np.logspace(np.log10(k[0]), np.log10(k[-1]),
                       4*Nk))
        kp = np.append(kp, np.logspace(np.log10(k[-1]), largek, Nk))

    Nz = 500
    z  = np.linspace(-1, 1, Nz)
    kij, pij, zij = np.meshgrid(kp, p, z, indexing='ij')
    ptilde2 = pij**2 + kij**2 - 2*kij*pij*zij
    ptilde  = np.sqrt(ptilde2)

    EK_p = np.interp(p, k, EK)
    if plot:
        plt.plot(p, EK_p)
        plt.xscale('log')
        plt.yscale('log')

    EK_ptilde = np.interp(ptilde, k, EK)
    ptilde[np.where(ptilde == 0)] = 1e-50

    Delta_mn = kij**0 - 1

    if terms == 'all':
        inds_m   = [-1, 1]
        inds_n   = [-1, 1]
        Delta_mn = np.zeros((4, len(kp), len(p), len(z)))
    tot_inds = 0
    l = 0
    for m in inds_m:
        for n in inds_n:
            Delta_mn[l, :, :, :] = \
                    compute_Delta_mn(tfin, kij*kstar, pij*kstar, ptilde*kstar,
                                     cs2=cs2, m=m, n=n, tini=tini, expansion=expansion)
            l += 1

    if l != 0: Delta_mn = Delta_mn/(l + 1)

    Omm = np.zeros((l + 1, len(kp)))
    for i in range(0, l):
        try:    Pi_1 = np.trapezoid(EK_ptilde/ptilde**4*(1 - zij**2)**2*Delta_mn[i, :, :, :],
                                    z, axis=2)
        except: Pi_1 = np.trapz(EK_ptilde/ptilde**4*(1 - zij**2)**2*Delta_mn[i, :, :, :],
                                z, axis=2)
        kij, EK_pij = np.meshgrid(kp, EK_p, indexing='ij')
        kij, pij    = np.meshgrid(kp, p, indexing='ij')
        try:    Omm[i, :]   = np.trapezoid(Pi_1*pij**2*EK_pij, p, axis=1)
        except: Omm[i, :]   = np.trapz(Pi_1*pij**2*EK_pij, p, axis=1)

    return kp, Omm

def compute_Delta_mn(t, k, p, ptilde, cs2=cs2_ref, m=1, n=1, tini=1.,
                     expansion=True):

    '''
    Function that computes the integrated Green's functions and the stationary
    UETC 4 Delta_mn used in the sound shell model for the computation of the GW
    spectrum.

    Reference: RoperPol:2023dzg, eqs.56-59

    Arguments:
        t      -- time
        k      -- wave number k
        p      -- wave number p to be integrated over
        ptilde -- second wave number tilde p to be integrated over
        cs2    -- square of the speed of sound (default is 1/3)
        m, n   -- indices of the Deltamn function
        tini   -- initial time of GW production in units of 1/Hubble
        expansion -- option to include the effect of the expansion of the Universe
                     during radiation domination (default True)
    '''

    cs = np.sqrt(cs2)
    pp = n*k + cs*(m*ptilde + p)
    pp[np.where(pp == 0)] = 1e-50

    if expansion:

        import scipy.special as spe
        si_t, ci_t       = spe.sici(pp*t)
        si_tini, ci_tini = spe.sici(pp*tini)

        # compute Delta Ci^2 and Delta Si^2
        DCi = ci_t - ci_tini
        DSi = si_t - si_tini

        Delta_mn = DCi**2 + DSi**2

    else:

        Delta_mn = 2*(1 - np.cos(pp*(t - tini)))/pp**2

    return Delta_mn

########### FUNCTIONS FOR THE LOCALLY STATIONARY UETC ###########

def K2int(dtfin, K0=1., dt0=dt0_ref, b=0., expansion=False, beta=100):

    '''
    Function that returns the integrated kinetic energy density
    squared for a power law

      K(dt) = K0 (dt/dt0)^(-b),

    where dt and dt0 represent time intervals, K0 is the amplitude
    at the end of the phase transition, and dt0 = 11 is a numerical
    parameter of the fit.

    Based on the results of Higgsless simulations of Caprini:2024gyk,
    see equation 2.27.

    It determines the amplitude of GWs in the locally stationary
    UETC describing the decay of compressional sources, see equation
    2.33

    Arguments:
        dtfin -- duration of the sourcing (in units of 1/beta or 1/H* for
                 expansion = False or True)
        K0    -- amplitude of K at the end of the phase transition
        dt0   -- numerical parameter used for the fits (= 11)
        b     -- power law exponent of the decay of K in time
        expansion -- option to consider flat Minkowski space-time
                     (if expansion is False, default) or a radiation-
                     dominated expanding Universe
        beta  -- nucleation rate normalized to the Hubble time,
                 beta/H*, relevant if expansion is chosen

    Returns:
        K2int -- integrated squared kinetic energy density
    '''

    K2int = K0**2*dt0

    # computation in Minkowski space-time
    if not expansion:
        if b == 0.5:
            K2int *= np.log(1 + dtfin/dt0)
        else:
            K2int *= ((1 + dtfin/dt0)**(1. - 2*b) - 1)/(1. - 2.*b)

    # computation in expanding Universe during radiation-domination
    else:

        import scipy.special as spe

        if beta < dt0:
            print('The value of beta in K2int cannot be smaller than',
                  'dt0 = %.1f for the chosen model'%dt0)
            print('Choose a larger value of beta to include expansion')
            return 0

        dt0   =  dt0/beta
        K2int *= 1./beta/(dt0 - 1.)**2

        if b == 0.5:
            K2int *= (dt0 - 1.)/(1. + dtfin) + np.log((1 + dtfin/dt0)/(1 + dtfin))

        else:
            K2int *= 1./(1. - 2*b)
            A      = spe.hyp2f1(2, 1 - 2*b, 2. - 2*b, (dt0 + dtfin)/(dt0 - 1))
            B      = spe.hyp2f1(2, 1 - 2*b, 2. - 2*b, dt0/(dt0 - 1.))
            K2int *= (1 + dtfin/dt0)**(1. - 2*b)*A - B

    return K2int
